@@ -1,9 +1,9 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-1999  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
+    Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
 
-	Andrew Clausen			<clausen@gnu.org>
-	Lennert Buytenhek		<buytenh@gnu.org>
+	Andrew Clausen			<clausen@alphalink.com.au>
+	Lennert Buytenhek		<buytenh@scotty.dsv.nl>
 	Matt Wilson, Red Hat Inc.	<msw@redhat.com>
 
     This program is free software; you can redistribute it and/or modify
@@ -24,7 +24,7 @@
 #ifndef PED_DEVICE_H_INCLUDED
 #define PED_DEVICE_H_INCLUDED
 
-#define PED_SECTOR_SIZE		512	/* FIXME */
+#define PED_SECTOR_SIZE		512
 typedef long long PedSector;
 
 typedef enum {
