@@ -1,9 +1,9 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-1999  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
+    Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
 
-	Andrew Clausen			<clausen@gnu.org>
-	Lennert Buytenhek		<buytenh@gnu.org>
+	Andrew Clausen			<clausen@alphalink.com.au>
+	Lennert Buytenhek		<buytenh@scotty.dsv.nl>
         Matt Wilson, Red Hat Inc.       <msw@redhat.com>
 
     This program is free software; you can redistribute it and/or modify
@@ -24,62 +24,6 @@
 #ifndef PED_DISK_H_INCLUDED
 #define PED_DISK_H_INCLUDED
 
-/* TODO:  remove these!  This stuff should be delegated away to
- * ped_file_system_get_part_id()
- */
-
-#define PED_FREE_SPACE 0x00
-
-#define PED_LINUX_EXTENDED 0x85
-#define PED_LINUX_MINIX 0x81
-#define PED_LINUX_SWAP 0x82
-#define PED_LINUX 0x83
-
-#define PED_DOS_EXTENDED 0x05
-#define PED_DOS_FAT12 0x01
-#define PED_DOS_FAT16 0x04
-#define PED_DOS_FAT16_BIG 0x06
-#define PED_DOS_HIDDEN_FAT12 0x11
-#define PED_DOS_HIDDEN_FAT16 0x14
-#define PED_DOS_HIDDEN_FAT16_BIG 0x16
-#define PED_DOS_ACCESS 0xe1
-#define PED_DOS_RO 0xe3
-#define PED_DOS_SECONDARY 0xf2
-
-#define PED_WIN95_EXTENDED 0x0f
-#define PED_WIN95_FAT32 0x0b
-#define PED_WIN95_FAT32_LBA 0x0c
-#define PED_WIN95_FAT16_LBA 0x0e
-#define PED_WIN95_HIDDEN_FAT32 0x1b
-#define PED_WIN95_HIDDEN_FAT32_LBA 0x1c
-#define PED_WIN95_HIDDEN_FAT16_LBA 0x1e
-
-#define PED_OS2_HPFS 0x07
-#define PED_OS2_BOOT 0x0a
-
-#define PED_GNU_HURD 0x63
-#define PED_MINIX_OLD 0x80
-#define PED_BSD_386 0xa5
-#define PED_NOVELL_NETWARE_286 0x64
-#define PED_NOVELL_NETWARE_386 0x65
-#define PED_NOVELL 0x51
-
-#define PED_XENIX_ROOT 0x02
-#define PED_XENIX_USR 0x03
-#define PED_AIX 0x08
-#define PED_AIX_BOOT 0x09
-#define PED_VENIX 0x40
-
-#define PED_MICROPORT 0x52
-#define PED_PC_IX 0x75
-#define PED_AMOEBA 0x93
-#define PED_AMOEBA_BBT 0x94
-#define PED_BSDI_FS 0xb7
-#define PED_BSDI_SWAP 0xb8
-#define PED_SYRINX 0xC7
-#define PED_CPM 0xdb
-#define PED_BBT 0xff
-
 typedef struct _PedGeometry		PedGeometry;
 typedef struct _PedDisk			PedDisk;
 typedef struct _PedPartition		PedPartition;
@@ -117,8 +61,8 @@ struct _PedPartition {
 	PedPartitionType	type;
 	PedPartition*		part_list;
 
-	unsigned char		system;
-	unsigned char		bootable;
+	int			system;
+	int			bootable;
 
 	PedFileSystemType*	fs_type;
 };
@@ -127,6 +71,9 @@ struct _PedDisk {
 	PedDevice*		dev;
 	PedDiskType*		type;
 	PedPartition*		part_list;
+
+/* office use only ;-) */
+	int			update_mode;
 };
 
 /* align_geometry() should only be called on partitions that have geometry,
@@ -146,12 +93,12 @@ struct _PedDiskOps {
 	int (*align_partition) (PedDisk* disk, PedPartition* part);
 	int (*enumerate_partition) (PedDisk* disk, PedPartition* part);
 	int (*alloc_metadata) (PedDisk* disk);
-	char (*get_extended_system) ();
+	int (*get_extended_system) ();
 };
 
 struct _PedDiskType {
 	PedDiskType*		next;
-	const char* const	name;
+	const char*		name;
 	PedDiskOps* const	ops;
 };
 
@@ -163,6 +110,7 @@ extern PedDiskType* ped_disk_probe (const PedDevice* dev);
 
 extern PedDisk* ped_disk_open (PedDevice* dev);
 extern PedDisk* ped_disk_create (PedDevice* dev, PedDiskType* type);
+extern PedDisk* ped_disk_alloc (PedDevice* dev, PedDiskType* type);
 extern int ped_disk_close (PedDisk* disk);
 extern int ped_disk_read (PedDisk* disk);
 extern int ped_disk_write (PedDisk* disk);
@@ -172,6 +120,13 @@ extern PedGeometry* ped_geometry_new (const PedDisk* disk, PedSector start,
 				      PedSector length);
 extern PedGeometry* ped_geometry_duplicate (const PedGeometry* geom);
 extern void ped_geometry_destroy (PedGeometry* geom);
+extern void ped_geometry_set (PedGeometry* geom, PedSector start,
+			      PedSector length);
+extern void ped_geometry_set_start (PedGeometry* geom, PedSector start);
+extern void ped_geometry_set_end (PedGeometry* geom, PedSector end);
+extern int ped_geometry_test_overlap (PedGeometry* a, PedGeometry* b);
+extern int ped_geometry_test_inside (PedGeometry* a, PedGeometry* b);
+
 extern int ped_geometry_read (const PedGeometry* geom, void* buffer,
 			      PedSector offset, PedSector count);
 extern int ped_geometry_write (PedGeometry* geom, const void* buffer,
