@@ -1,9 +1,9 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-1999  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
+    Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
 
-	Andrew Clausen			<clausen@gnu.org>
-	Lennert Buytenhek		<buytenh@gnu.org>
+	Andrew Clausen			<clausen@alphalink.com.au>
+	Lennert Buytenhek		<buytenh@scotty.dsv.nl>
 	Matt Wilson, Red Hat Inc.	<msw@redhat.com>
 
     This program is free software; you can redistribute it and/or modify
@@ -38,9 +38,12 @@ extern "C" {
 #include <parted/exception.h>
 #include <parted/filesys.h>
 
+#include <stdlib.h>
+
 extern int ped_init ();
 extern void ped_done ();
-extern void* ped_malloc (int size);
+extern void* ped_malloc (size_t size);
+extern int ped_realloc (void** ptr, size_t size);
 extern void ped_free (void* ptr);
 
 #ifdef __cplusplus
