@@ -33,11 +33,11 @@
 
 #include "llseek.h"
 
-#include <string.h>
-#include <fcntl.h>
 #include <errno.h>
-#include <unistd.h>
+#include <fcntl.h>
 #include <string.h>
+#include <string.h>
+#include <unistd.h>
 #include <linux/hdreg.h>
 #include <linux/major.h>
 #include <linux/fs.h>		/* FIXME: autoconfuse this */
@@ -148,24 +148,34 @@ strip_name(char* n)
 static int
 device_probe_type (PedDevice* dev)
 {
-	struct stat	dev_stat;
-	int		dev_major;
+	struct stat		dev_stat;
+	int			dev_major;
+	PedExceptionOption	ex_status;
 
 	if (!ped_device_stat (dev, &dev_stat))
 		return 0;
 
 	if (!S_ISBLK(dev_stat.st_mode)) {
-		ped_exception_throw (PED_EXCEPTION_WARNING,
-				     PED_EXCEPTION_CANCEL,
-				     _("%s is not a block device."),
-				     dev->path);
-		dev->type = PED_DEVICE_UNKNOWN;
-		return 1;
+		ex_status = ped_exception_throw (
+				PED_EXCEPTION_WARNING,
+				PED_EXCEPTION_IGNORE_CANCEL,
+				_("%s is not a block device."),
+				dev->path);
+		switch (ex_status) {
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_IGNORE:
+				dev->type = PED_DEVICE_UNKNOWN;
+				return 1;
+		}
 	}
 
 	dev_major = major (dev_stat.st_rdev);
 
-	if (SCSI_DISK_MAJOR (dev_major) && ((dev_stat.st_rdev & 0x0f) == 0)) {
+	if (SCSI_BLK_MAJOR (dev_major) && ((dev_stat.st_rdev & 0x0f) == 0)) {
 		dev->type = PED_DEVICE_SCSI;
 		return 1;
 	}
@@ -177,15 +187,21 @@ device_probe_type (PedDevice* dev)
 		return 1;
 	}
 
-	if (ped_exception_throw (PED_EXCEPTION_WARNING,
-				 PED_EXCEPTION_IGNORE_CANCEL,
-				 _("Device %s is neither a SCSI nor IDE "
-				   "drive."),
-				  dev->path)
-			== PED_EXCEPTION_CANCEL)
-		return 0;
-	dev->type = PED_DEVICE_UNKNOWN;
-	return 1;
+	ex_status = ped_exception_throw (
+			PED_EXCEPTION_WARNING,
+			PED_EXCEPTION_IGNORE_CANCEL,
+			_("Device %s is neither a SCSI nor IDE drive."),
+			dev->path);
+	switch (ex_status) {
+		case PED_EXCEPTION_CANCEL:
+			return 0;
+
+		case PED_EXCEPTION_UNHANDLED:
+			ped_exception_catch ();
+		case PED_EXCEPTION_IGNORE:
+			dev->type = PED_DEVICE_UNKNOWN;
+			return 1;
+	}
 }
 
 static int
@@ -207,7 +223,7 @@ device_get_sector_size (PedDevice* dev)
 			"not to work properly with drives with sector sizes "
 			"other than 512 bytes"),
 			dev->path,
-			dev->sector_size)
+			sector_size)
 				== PED_EXCEPTION_IGNORE)
 			return sector_size;
 		else
@@ -253,7 +269,7 @@ device_probe_geometry (PedDevice* dev)
 		goto error;
 
 	if (ioctl (dev->fd, HDIO_GETGEO, &geometry)) {
-		ped_exception_throw (PED_EXCEPTION_FATAL,
+		ped_exception_throw (PED_EXCEPTION_ERROR,
 				     PED_EXCEPTION_CANCEL,
 				     _("Could not read geometry of %s - %s."),
 				     dev->path, strerror (errno));
@@ -290,22 +306,31 @@ init_ide (PedDevice* dev)
 	struct stat		dev_stat;
 	int			dev_major;
 	struct hd_driveid	hdi;
+	PedExceptionOption	ex_status;
 
 	if (!ped_device_stat (dev, &dev_stat))
 		goto error;
+
 	dev_major = major (dev_stat.st_rdev);
 
 	if (!ped_device_open (dev))
 		goto error;
 
         if (ioctl (dev->fd, HDIO_GET_IDENTITY, &hdi)) {
-		if (ped_exception_throw (PED_EXCEPTION_WARNING,
-			     PED_EXCEPTION_IGNORE_CANCEL,
-			     _("Could not get identity of device %s - %s"),
-			     dev->path, strerror (errno))
-		    == PED_EXCEPTION_CANCEL)
-			goto error_close_dev;
-		dev->model = strdup(_("unknown"));
+		ex_status = ped_exception_throw (
+				PED_EXCEPTION_WARNING,
+				PED_EXCEPTION_IGNORE_CANCEL,
+				_("Could not get identity of device %s - %s"),
+				dev->path, strerror (errno));
+		switch (ex_status) {
+			case PED_EXCEPTION_CANCEL:
+				goto error_close_dev;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_IGNORE:
+				dev->model = strdup(_("unknown"));
+		}
 	} else {
 		dev->model = strip_name (hdi.model);
 	}
@@ -313,10 +338,6 @@ init_ide (PedDevice* dev)
 	if (!device_probe_geometry (dev))
 		goto error_close_dev;
 
-	dev->host = (dev_major == IDE1_MAJOR) + (dev_major == IDE2_MAJOR) * 2
-		    + (dev_major == IDE3_MAJOR) * 3;
-	dev->did = (dev_stat.st_rdev & 0x40) >> 6;
-
 	ped_device_close (dev);
 	return 1;
 
@@ -358,8 +379,8 @@ init_scsi (PedDevice* dev)
 		dev->model = _("Unknown SCSI");
 	}
 
-	if (ioctl(dev->fd, SCSI_IOCTL_GET_IDLUN, idlun)) {
-			ped_exception_throw (PED_EXCEPTION_FATAL,
+	if (ioctl (dev->fd, SCSI_IOCTL_GET_IDLUN, idlun)) {
+			ped_exception_throw (PED_EXCEPTION_ERROR,
 				     PED_EXCEPTION_CANCEL,
 				     _("Error initialising SCSI device "
 				       "%s - %s"),
@@ -387,6 +408,7 @@ init_unknown (PedDevice* dev)
 {
 	struct stat		dev_stat;
 	struct hd_geometry	geometry;
+	PedExceptionOption	ex_status;
 
 	if (!ped_device_stat (dev, &dev_stat))
 		goto error;
@@ -402,19 +424,27 @@ init_unknown (PedDevice* dev)
 		ped_exception_catch ();
 		ped_exception_leave_all ();
 
-		dev->length = dev_stat.st_size / 512;
-		if (ped_exception_throw (PED_EXCEPTION_WARNING,
+		ex_status = ped_exception_throw (
+				PED_EXCEPTION_WARNING,
 				PED_EXCEPTION_IGNORE_CANCEL,
 				_("Unable to determine geometry of "
 				"file/device.  You should not use Parted "
-				"unless you REALLY know what you're doing!"))
-				!= PED_EXCEPTION_IGNORE)
-			goto error;
+				"unless you REALLY know what you're doing!"));
+		switch (ex_status) {
+			case PED_EXCEPTION_CANCEL:
+				goto error_close_dev;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_IGNORE:
+		}
 
 		/* what should we stick in here? */
+		dev->length = dev_stat.st_size / 512;
 		dev->cylinders = dev->length / 4 / 32;
 		dev->heads = 4;
 		dev->sectors = 32;
+		dev->sector_size = 512;
 	}
 
 	ped_device_close (dev);
@@ -497,8 +527,11 @@ ped_device_open (PedDevice* dev)
 
 	dev->fd = open (dev->path, O_RDWR);
 	if (dev->fd == -1) {
-		ped_exception_throw (PED_EXCEPTION_FATAL, PED_EXCEPTION_CANCEL,
-				     "%s: %s", dev->path, strerror (errno));
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			"Error opening %s: %s",
+			dev->path, strerror (errno));
 		return 0;
 	}
 
@@ -508,7 +541,9 @@ ped_device_open (PedDevice* dev)
 int
 ped_device_close (PedDevice* dev)
 {
-	int	retry_count = 5;	/* a bad guess */
+	int			retry_count = 5;	/* a bad guess */
+	PedExceptionOption	ex_status;
+
 	PED_ASSERT (dev != NULL, return 0);
 
 	if (--dev->open_count) return 1;
@@ -520,16 +555,19 @@ ped_device_close (PedDevice* dev)
 			retry_count--;
 			sync();
 	    		if (!retry_count) {
-				ped_exception_throw (
+				ex_status = ped_exception_throw (
 					PED_EXCEPTION_WARNING,
-					PED_EXCEPTION_OK,
+					PED_EXCEPTION_IGNORE,
 		_("The kernel was unable to re-read the partition "
 		  "table on %s (%s).  This means Linux knows nothing "
 		  "about any modifications you made.  You should "
 		  "reboot your computer before doing anything with %s."),
 					dev->path, strerror (errno), dev->path);
 
-	    			break;
+				if (ex_status == PED_EXCEPTION_UNHANDLED)
+					ped_exception_catch ();
+
+				break;
 	    		}
 		}
 	}
@@ -544,7 +582,7 @@ ped_device_seek (const PedDevice* dev, PedSector sector)
 {
 	PED_ASSERT (dev != NULL, return 0);
 
-#ifdef linux
+#if SIZEOF_OFF_T < 8 && defined(linux)
 	if (sizeof (off_t) < 8) {
 		loff_t		pos = sector * 512;
 		return ped_llseek (dev->fd, pos, SEEK_SET) == pos;
@@ -560,8 +598,9 @@ int
 ped_device_read (const PedDevice* dev, void* buffer, PedSector start,
 		 PedSector count)
 {
-	int		status;
-	int		exception_status;
+	int			status;
+	PedExceptionOption	ex_status;
+	size_t			read_length = count * PED_SECTOR_SIZE;
 
 	PED_ASSERT (dev != NULL, return 0);
 	PED_ASSERT (buffer != NULL, return 0);
@@ -570,27 +609,54 @@ ped_device_read (const PedDevice* dev, void* buffer, PedSector start,
 		if (ped_device_seek (dev, start))
 			break;
 
-		exception_status = ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
+		ex_status = ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_RETRY_IGNORE_CANCEL,
 			_("%s during seek for read on %s"),
 			strerror (errno), dev->path);
 
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY) return 0;
+		switch (ex_status) {
+			case PED_EXCEPTION_IGNORE:
+				return 1;
+
+			case PED_EXCEPTION_RETRY:
+				break;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+		}
 	}
 
 	while (1) {
-		status = read (dev->fd, buffer, count * PED_SECTOR_SIZE);
+		status = read (dev->fd, buffer, read_length);
 		if (status == count * PED_SECTOR_SIZE) break;
+		if (status > 0) {
+			read_length -= status;
+			buffer += status;
+			continue;
+		}
 
-		exception_status = ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
+		ex_status = ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_RETRY_IGNORE_CANCEL,
 			_("%s during read on %s"),
 			strerror (errno),
 			dev->path);
 
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY) return 0;
+		switch (ex_status) {
+			case PED_EXCEPTION_IGNORE:
+				return 1;
+
+			case PED_EXCEPTION_RETRY:
+				break;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+		}
 	}
 
 	return 1;
@@ -599,8 +665,8 @@ ped_device_read (const PedDevice* dev, void* buffer, PedSector start,
 int
 ped_device_sync (PedDevice* dev)
 {
-	int		status;
-	int		exception_status;
+	int			status;
+	PedExceptionOption	ex_status;
 
 	PED_ASSERT (dev != NULL, return 0);
 
@@ -608,13 +674,24 @@ ped_device_sync (PedDevice* dev)
 		status = fsync (dev->fd);
 		if (status >= 0) break;
 
-		exception_status = ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
+		ex_status = ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_RETRY_IGNORE_CANCEL,
 			_("%s during write on %s"),
 			strerror (errno), dev->path);
 
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY) return 0;
+		switch (ex_status) {
+			case PED_EXCEPTION_IGNORE:
+				return 1;
+
+			case PED_EXCEPTION_RETRY:
+				break;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+		}
 	} 
 
 	return 1;
@@ -624,8 +701,9 @@ int
 ped_device_write (PedDevice* dev, const void* buffer, PedSector start,
 		  PedSector count)
 {
-	int		status;
-	int		exception_status;
+	int			status;
+	PedExceptionOption	ex_status;
+	size_t			write_length = count * PED_SECTOR_SIZE;
 
 	PED_ASSERT (dev != NULL, return 0);
 	PED_ASSERT (buffer != NULL, return 0);
@@ -634,13 +712,23 @@ ped_device_write (PedDevice* dev, const void* buffer, PedSector start,
 		if (ped_device_seek (dev, start))
 			break;
 
-		exception_status = ped_exception_throw (
+		ex_status = ped_exception_throw (
 			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
 			_("%s during seek for write on %s"),
 			strerror (errno), dev->path);
 
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY) return 0;
+		switch (ex_status) {
+			case PED_EXCEPTION_IGNORE:
+				return 1;
+
+			case PED_EXCEPTION_RETRY:
+				break;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+		}
 	}
 
 #ifdef READONLY
@@ -649,16 +737,32 @@ ped_device_write (PedDevice* dev, const void* buffer, PedSector start,
 #else
 	dev->dirty = 1;
 	while (1) {
-		status = write (dev->fd, buffer, count * PED_SECTOR_SIZE);
+		status = write (dev->fd, buffer, write_length);
 		if (status == count * PED_SECTOR_SIZE) break;
+		if (status > 0) {
+			write_length -= status;
+			buffer += status;
+			continue;
+		}
 
-		exception_status = ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
+		ex_status = ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_RETRY_IGNORE_CANCEL,
 			_("%s during write on %s"),
 			strerror (errno), dev->path);
 
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY) return 0;
+		switch (ex_status) {
+			case PED_EXCEPTION_IGNORE:
+				return 1;
+
+			case PED_EXCEPTION_RETRY:
+				break;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+		}
 	}
 #endif
 	return 1;
@@ -670,38 +774,25 @@ PedSector
 ped_device_check (PedDevice* dev, void* buffer, PedSector start,
 		  PedSector count)
 {
-	int		status;
-	int		exception_status;
+	int			status;
+	int			done = 0;
 
 	PED_ASSERT (dev != NULL, return 0);
 	PED_ASSERT (buffer != NULL, return 0);
 
+	if (!ped_device_seek (dev, start))
+		return 0;
+
 	while (1) {
-		if (ped_device_seek (dev, start))
+		status = read (dev->fd, buffer, (count-done) * PED_SECTOR_SIZE);
+		if (status < 0)
+			return 0;
+
+		done += status / PED_SECTOR_SIZE;
+		if (done == count * PED_SECTOR_SIZE)
 			break;
-
-		exception_status = ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
-			_("%s during seek for check on %s"),
-			strerror (errno), dev->path);
-
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY) return 0;
-	}
-
-	while (1) {
-		status = read (dev->fd, buffer, count * PED_SECTOR_SIZE);
-		if (status == count * PED_SECTOR_SIZE) break;
-
-		exception_status = ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_RETRY_IGNORE_CANCEL,
-			_("%s during check on %s"),
-			strerror (errno),
-			dev->path);
-
-		if (exception_status == PED_EXCEPTION_IGNORE) break;
-		if (exception_status != PED_EXCEPTION_RETRY)
-			return status / PED_SECTOR_SIZE;
+		else
+			continue;
 	}
 
 	return count;
