@@ -1,4 +1,4 @@
-/*
+ /*
     libparted - a library for manipulating disk partitions
     Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
 
@@ -42,6 +42,13 @@
 #include <linux/fs.h>
 #endif
 
+/* UPDATE MODE functions */
+static int _disk_raw_insert_before (PedDisk* disk, PedPartition* loc,
+				    PedPartition* part);
+static int _disk_raw_insert_after (PedDisk* disk, PedPartition* loc,
+				   PedPartition* part);
+static int _disk_raw_remove (PedDisk* disk, PedPartition* part);
+
 static PedDiskType*	disk_types = NULL;
 
 void
@@ -100,11 +107,17 @@ ped_disk_probe (const PedDevice* dev)
 
 	PED_ASSERT (dev != NULL, return NULL);
 
+	ped_exception_leave_all ();
+
 	while (1) {
 		walk = ped_disk_type_get_next (walk);
 		if (!walk) break;
 		if (walk->ops->probe (dev)) break;
 	}
+
+	if (ped_exception)
+		ped_exception_catch ();
+	ped_exception_fetch_all ();
 	return walk;
 }
 
@@ -141,6 +154,23 @@ ped_disk_create (PedDevice* dev, PedDiskType* type)
 	return type->ops->create (dev);
 }
 
+PedDisk*
+ped_disk_alloc (PedDevice* dev, PedDiskType* disk_type)
+{
+	PedDisk*	disk;
+
+	disk = (PedDisk*) ped_malloc (sizeof (PedDisk));
+	if (!disk)
+		return NULL;
+
+	disk->dev = dev;
+	disk->type = disk_type;
+	disk->part_list = NULL;
+	disk->update_mode = 0;
+
+	return disk;
+}
+
 int
 ped_disk_close (PedDisk* disk)
 {
@@ -161,7 +191,9 @@ ped_disk_write (PedDisk* disk)
 	PED_ASSERT (disk != NULL, return 0);
 
 	if (!disk->type->ops->write) {
-		ped_exception_throw (PED_EXCEPTION_BUG, PED_EXCEPTION_CANCEL,
+		ped_exception_throw (
+			PED_EXCEPTION_NO_FEATURE,
+			PED_EXCEPTION_CANCEL,
 			_("The code to write the partition table hasn't been "
 			"written for %s yet"),
 			disk->type->name);
@@ -228,7 +260,10 @@ ped_disk_enumerate_partitions (PedDisk* disk)
 
 	PED_ASSERT (disk != NULL, return 0);
 
-/* first "sort" already-numbered partitions */
+/* first "sort" already-numbered partitions.  (Eg: if a logical partition
+ * is removed, then all logical partitions that were number higher MUST be
+ * renumbered)
+ */
 	end = ped_disk_last_partition_num (disk);
 	for (i=1; i<=end; i++) {
 		walk = ped_disk_get_partition (disk, i);
@@ -249,8 +284,8 @@ ped_disk_enumerate_partitions (PedDisk* disk)
 	return 1;
 }
 
-int
-ped_disk_remove_metadata (PedDisk* disk)
+static int
+_disk_remove_metadata (PedDisk* disk)
 {
 	PedPartition*	walk = NULL;
 	PedPartition*	next;
@@ -272,15 +307,154 @@ ped_disk_remove_metadata (PedDisk* disk)
 	return 1;
 }
 
-int
-ped_disk_alloc_metadata (PedDisk* disk)
+static int
+_disk_alloc_metadata (PedDisk* disk)
 {
 	PED_ASSERT (disk != NULL, return 0);
 
-	ped_disk_remove_metadata (disk);
+	if (!disk->update_mode)
+		_disk_remove_metadata (disk);
+
 	return disk->type->ops->alloc_metadata (disk);
 }
 
+static int
+_disk_remove_freespace (PedDisk* disk)
+{
+	PedPartition*	walk;
+	PedPartition*	next;
+
+	walk = ped_disk_next_partition (disk, NULL);
+	for (; walk; walk = next) {
+		next = ped_disk_next_partition (disk, walk);
+
+		if (walk->type & PED_PARTITION_FREESPACE)
+			_disk_raw_remove (disk, walk);
+	}
+
+	return 1;
+}
+
+static int
+_alloc_extended_freespace (PedDisk* disk)
+{
+	PedSector	last_end;
+	PedPartition*	walk;
+	PedPartition*	last;
+	PedPartition*	free_space;
+	PedPartition*	extended_part;
+
+	extended_part = ped_disk_extended_partition (disk);
+	if (!extended_part)
+		return 1;
+
+	last_end = extended_part->geom.start;
+	last = NULL;
+	
+	for (walk = extended_part->part_list; walk; walk = walk->next) {
+		if (walk->geom.start > last_end + 1) {
+			free_space = ped_partition_new (
+					disk,
+					PED_PARTITION_FREESPACE
+						| PED_PARTITION_LOGICAL,
+					NULL,
+					last_end + 1, walk->geom.start - 1);
+			_disk_raw_insert_before (disk, walk, free_space);
+		}
+
+		last = walk;
+		last_end = last->geom.end;
+	}
+
+	if (last_end < extended_part->geom.end) {
+		free_space = ped_partition_new (
+				disk,
+				PED_PARTITION_FREESPACE | PED_PARTITION_LOGICAL,
+				NULL,
+				last_end + 1, extended_part->geom.end);
+
+		if (last)
+			return _disk_raw_insert_after (disk, last, free_space);
+		else
+			extended_part->part_list = free_space;
+	}
+
+	return 1;
+}
+
+static int
+_disk_alloc_freespace (PedDisk* disk)
+{
+	PedSector	last_end;
+	PedPartition*	walk;
+	PedPartition*	last;
+	PedPartition*	free_space;
+
+	if (!_disk_remove_freespace (disk))
+		return 0;
+	if (!_alloc_extended_freespace (disk))
+		return 0;
+
+	last = NULL;
+	last_end = 0;
+
+	for (walk = disk->part_list; walk; walk = walk->next) {
+		if (walk->geom.start > last_end + 1) {
+			free_space = ped_partition_new (disk,
+					PED_PARTITION_FREESPACE, NULL,
+					last_end + 1, walk->geom.start - 1);
+			_disk_raw_insert_before (disk, walk, free_space);
+		}
+
+		last = walk;
+		last_end = last->geom.end;
+	}
+
+	if (last_end < disk->dev->length - 1) {
+		free_space = ped_partition_new (disk,
+					PED_PARTITION_FREESPACE, NULL,
+					last_end + 1, disk->dev->length - 1);
+		if (last)
+			return _disk_raw_insert_after (disk, last, free_space);
+		else
+			disk->part_list = free_space;
+	}
+
+	return 1;
+}
+
+/* update mode: used when updating the internal representation of the partition
+ * table.  In update mode, the metadata and freespace placeholder/virtual
+ * partitions are removed, making it much easier for various manipulation
+ * routines...
+ */
+static void
+_disk_push_update_mode (PedDisk* disk)
+{
+	if (!disk->update_mode) {
+		_disk_remove_freespace (disk);
+		disk->update_mode++;
+		_disk_remove_metadata (disk);
+	} else {
+		disk->update_mode++;
+	}
+}
+
+static void
+_disk_pop_update_mode (PedDisk* disk)
+{
+	if (disk->update_mode == 1) {
+	/* re-allocate metadata BEFORE leaving update mode, to prevent infinite
+	 * recursion (metadata allocation requires update mode)
+	 */
+		_disk_alloc_metadata (disk);
+		disk->update_mode--;
+		_disk_alloc_freespace (disk);
+	} else {
+		disk->update_mode--;
+	}
+}
+
 PedPartition*
 ped_partition_new (const PedDisk* disk, PedPartitionType type,
 		   PedFileSystemType* fs_type, PedSector start, PedSector end)
@@ -303,9 +477,7 @@ ped_partition_new (const PedDisk* disk, PedPartitionType type,
 	part->next = NULL;
 
 	part->geom.disk = (PedDisk*) disk;
-	part->geom.start = start;
-	part->geom.end = end;
-	part->geom.length = end - start + 1;
+	ped_geometry_set (&part->geom, start, end - start + 1);
 
 	part->type = type;
 	part->part_list = NULL;
@@ -670,15 +842,15 @@ ped_disk_next_partition (const PedDisk* disk, const PedPartition* part)
 	return NULL;
 }
 
-/* should be able to remove this */
 static int
-check_sanity (PedDisk* disk)
+_disk_check_sanity (PedDisk* disk)
 {
 	PedPartition*	walk;
 
 	PED_ASSERT (disk != NULL, return 0);
 
 	for (walk = disk->part_list; walk; walk = walk->next) {
+		PED_ASSERT (!(walk->type & PED_PARTITION_LOGICAL), return 0);
 		if (walk->prev)
 			PED_ASSERT (walk->prev->next == walk, return 0);
 	}
@@ -688,6 +860,7 @@ check_sanity (PedDisk* disk)
 
 	for (walk = ped_disk_extended_partition (disk)->part_list; walk;
 	     walk = walk->next) {
+		PED_ASSERT (walk->type & PED_PARTITION_LOGICAL, return 0);
 		if (walk->prev)
 			PED_ASSERT (walk->prev->next == walk, return 0);
 	}
@@ -730,15 +903,17 @@ ped_disk_get_partition_by_sector (const PedDisk* disk, PedSector sect)
 	}
 
 	/* should never get here, unless sect is outside of disk's useable
-	 * part.
+	 * part, or we're in "update mode", and the free space place-holders
+	 * have been removed with _disk_remove_freespace()
 	 */
 	return NULL;
 }
 
 /* I'm beginning to agree with Sedgewick :-/
+ * UPDATE MODE ONLY
  */
 static int
-disk_raw_insert_before (PedDisk* disk, PedPartition* loc, PedPartition* part)
+_disk_raw_insert_before (PedDisk* disk, PedPartition* loc, PedPartition* part)
 {
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (loc != NULL, return 0);
@@ -760,9 +935,10 @@ disk_raw_insert_before (PedDisk* disk, PedPartition* loc, PedPartition* part)
 }
 
 /* inserts a partition *after* loc.
+ * UPDATE MODE ONLY
  */
 static int
-disk_raw_insert_after (PedDisk* disk, PedPartition* loc, PedPartition* part)
+_disk_raw_insert_after (PedDisk* disk, PedPartition* loc, PedPartition* part)
 {
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (loc != NULL, return 0);
@@ -777,8 +953,10 @@ disk_raw_insert_after (PedDisk* disk, PedPartition* loc, PedPartition* part)
 	return 1;
 }
 
+/* UPDATE MODE ONLY
+ */
 static int
-disk_raw_remove (PedDisk* disk, PedPartition* part)
+_disk_raw_remove (PedDisk* disk, PedPartition* part)
 {
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (part != NULL, return 0);
@@ -788,147 +966,76 @@ disk_raw_remove (PedDisk* disk, PedPartition* part)
 		if (part->next)
 			part->next->prev = part->prev;
 	} else {
-		if (disk->part_list == part) {
-			disk->part_list = part->next;
-		} else {
+		if (part->type & PED_PARTITION_LOGICAL) {
 			ped_disk_extended_partition (disk)->part_list
 				= part->next;
+		} else {
+			disk->part_list = part->next;
 		}
 		if (part->next)
 			part->next->prev = NULL;
 	}
 
-/* for debugging */
-	part->next = NULL;
-	part->prev = NULL;
-
 	return 1;
 }
 
-/* chops the end of a partition off *after* sector "sect", and makes the
- * chopped off bit (to the right) free space.  Returns the free space.
+/* returns 0 if the partition, "part" overlaps with any partitions on the
+ * "disk".  The geometry of "part" is taken to be "geom", NOT "part->geom"
+ * (the idea here is to check if "geom" is valid, before changing "part").
+ * 
+ * This is useful for seeing if a resized partitions new geometry is going to
+ * fit, without the existing geomtry getting in the way.
+ *
+ * Note: overlap with an extended partition is also allowed, provided that
+ * "geom" lies completely inside the extended partition.
  */
-static PedPartition*
-ped_disk_split_right (PedDisk* disk, PedPartition* part, PedSector sect)
+int
+ped_disk_check_overlap_new_geom (PedDisk* disk, PedPartition* part,
+				 PedGeometry* geom)
 {
-	PedPartition*		tmp;
-	PedPartitionType	type;
+	PedPartition*	walk;
 
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (part != NULL, return 0);
-
-	type = (part->type & PED_PARTITION_LOGICAL) | PED_PARTITION_FREESPACE;
-
-	if (sect == part->geom.start - 1) {
-		part->type = type;
-		return part;
-	}
-	if (sect < part->geom.start || sect >= part->geom.end)
-		return NULL;
-
-	tmp = ped_partition_new (disk, type, NULL, sect + 1, part->geom.end);
-	if (!disk_raw_insert_after (disk, part, tmp))
-		return NULL;
-
-	part->geom.end = sect;
-	part->geom.length = part->geom.end - part->geom.start + 1;
-
-	return tmp;
-}
-
-/* chops the beginning of partition off and returns the chopped off bit.
- * ie: the right bit is preserved as "part", the left bit is returned as
- * free space.
- */
-static PedPartition*
-ped_disk_split_left (PedDisk* disk, PedPartition* part, PedSector sect)
-{
-	PedPartition*		tmp;
-	PedPartitionType	type;
-
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (part != NULL, return 0);
-
-	type = (part->type & PED_PARTITION_LOGICAL) | PED_PARTITION_FREESPACE;
-
-	if (sect == part->geom.end + 1) {
-		part->type = type;
-		return part;
-	}
-	if (sect < part->geom.start || sect > part->geom.end)
-		return NULL;
-
-	tmp = ped_partition_new (disk, type, NULL, part->geom.start, sect - 1);
-	if (!disk_raw_insert_before (disk, part, tmp))
-		return NULL;
-
-	part->geom.start = sect;
-	part->geom.length = part->geom.end - part->geom.start + 1;
-
-	return tmp;
-}
-
-static int
-ped_disk_replace_partition (PedDisk* disk, PedPartition* old, PedPartition* new)
-{
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (old != NULL, return 0);
-	PED_ASSERT (new != NULL, return 0);
-
-	if (!disk_raw_insert_before (disk, old, new))
-		return 0;
-	if (!disk_raw_remove (disk, old))
-		return 0;
-	ped_partition_destroy (old);
-
-	return 1;
-}
-
-static int
-ped_disk_join_freespace (PedDisk* disk)
-{
-	PedPartition*		walk;
-	PedPartition*		next;
-
-	PED_ASSERT (disk != NULL, return 0);
-
-	for (walk = disk->part_list; walk;
+	for (walk = ped_disk_next_partition (disk, NULL); walk;
 	     walk = ped_disk_next_partition (disk, walk)) {
-		next = walk->next;
-		if (!next)
+		if (walk->type & PED_PARTITION_FREESPACE)
 			continue;
-		if (walk->type & PED_PARTITION_FREESPACE
-		    && next->type & PED_PARTITION_FREESPACE) {
-			walk->geom.end = next->geom.end;
-			walk->geom.length
-				= walk->geom.end - walk->geom.start + 1;
-			disk_raw_remove (disk, next);
+		if (walk == part)
+			continue;
+		if (part->type & PED_PARTITION_EXTENDED
+		    && walk->type & PED_PARTITION_LOGICAL)
+			continue;
+
+		if (ped_geometry_test_overlap (&walk->geom, geom)) {
+			if (walk->type & PED_PARTITION_EXTENDED
+			    && part->type & PED_PARTITION_LOGICAL
+			    && ped_geometry_test_inside (&walk->geom, geom))
+				continue;
+			return 0;
 		}
 	}
-	check_sanity (disk);
 
 	return 1;
 }
 
 int
-ped_disk_add_partition (PedDisk* disk, PedPartition* part)
+ped_disk_check_overlap (PedDisk* disk, PedPartition* part)
 {
-	PedPartition*		free_left;
-	PedPartition*		free_required;
-	PedPartition*		free_right;
-	PedPartition*		ext_part;
+	return ped_disk_check_overlap_new_geom (disk, part, &part->geom);
+}
 
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (part != NULL, return 0);
+static int
+_check_new_partition_basic_sanity (PedDisk* disk, PedPartition* part)
+{
+	PedPartition*	ext_part = ped_disk_extended_partition (disk);
 
-	ext_part = ped_disk_extended_partition (disk);
-
-#ifdef VERBOSE
-	printf ("ped_disk_add_partition (dev=\"%s\", start=%d, end=%d,"
-		" type=%x)\n",
-		disk->dev->path, (int) part->geom.start, (int) part->geom.end,
-		(int) part->system);
-#endif
+	if (part->geom.start < 0 || part->geom.end >= disk->dev->length) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("Can't create a partition outside of the device."),
+			disk->dev->path);
+		return 0;
+	}
 
 	if ((part->type & PED_PARTITION_LOGICAL) && !ext_part) {
 		ped_exception_throw (
@@ -940,94 +1047,128 @@ ped_disk_add_partition (PedDisk* disk, PedPartition* part)
 		return 0;
 	}
 
-	if (ped_partition_is_active (part)) {
-		if (!ped_disk_enumerate_partition (disk, part))
-			return 0;
-		if (!ped_disk_align_partition (disk, part))
-			return 0;
-	}
+	return 1;
+}
 
-	if (part->type == PED_PARTITION_EXTENDED) {
-		if (ext_part) {
-			ped_exception_throw (
-				PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_CANCEL,
-				_("Can't have more than one extended partition "
-				"on %s"),
-				disk->dev->path);
-			return 0;
-		}
-		part->part_list = ped_partition_new (
-			disk, PED_PARTITION_LOGICAL | PED_PARTITION_FREESPACE,
-			NULL, part->geom.start, part->geom.end);
-	}
+static int
+_check_new_partition (PedDisk* disk, PedPartition* part)
+{
+	PedPartition*	ext_part = ped_disk_extended_partition (disk);
 
-	if (part->type & PED_PARTITION_LOGICAL) {
-		if (part->geom.start < ext_part->geom.start
-		    || part->geom.end > ext_part->geom.end) {
-			ped_exception_throw (
-				PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_CANCEL,
-				_("Can't add the logical partition outside "
-				"of the extended partition on %s."),
-				disk->dev->path);
-			return 0;
-		}
-	}
-
-	free_left = ped_disk_get_partition_by_sector (disk, part->geom.start);
-	if (!free_left
-	    || !(free_left->type & PED_PARTITION_FREESPACE)) {
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-			_("The start of the new partition doesn't lie on free "
-			  "space."));
+	if (part->type == PED_PARTITION_EXTENDED && ext_part) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("Can't have more than one extended partition on %s"),
+			disk->dev->path);
 		return 0;
 	}
+
+	if (part->type & PED_PARTITION_LOGICAL
+	    && !ped_geometry_test_inside (&ext_part->geom, &part->geom)) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("Can't add a logical partition outside of the "
+			  "extended partition on %s."),
+			disk->dev->path);
+		return 0;
+	}
+
+	if (!ped_disk_check_overlap (disk, part)) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("The new partition overlaps with another "
+			  "partition."));
+		return 0;
+	}
+
 	if (part->type == PED_PARTITION_PRIMARY
-	    && free_left->type & PED_PARTITION_LOGICAL) {
+	    && ext_part
+	    && ped_geometry_test_inside (&ext_part->geom, &part->geom)) {
 		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
 			_("Can't add a primary partition inside an extended "
 			 "partition."));
 		return 0;
 	}
-	if (part->geom.end > free_left->geom.end) {
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-			_("The end of the new partition doesn't lie on free "
-			  "space, or overlaps with another partition."));
-		return 0;
-	}
-	if (free_left->geom.start < part->geom.start)
-		free_required = ped_disk_split_right (disk, free_left,
-						      part->geom.start - 1);
-	else
-		free_required = free_left;
 
-	if (free_required->geom.end > part->geom.end)
-		free_right = ped_disk_split_right (disk, free_required,
-						   part->geom.end);
-	else
-		free_right = NULL;
-
-	if (!ped_disk_replace_partition (disk, free_required, part))
-		return 0;
-	if (!ped_disk_join_freespace (disk))
-		return 0;
-
-	if (!(part->type & PED_PARTITION_METADATA)) {
-		if (!ped_disk_alloc_metadata (disk)) {
-			free_required = ped_partition_duplicate (part);
-			free_required->type |= PED_PARTITION_FREESPACE;
-			disk_raw_insert_before (disk, part, free_required);
-			disk_raw_remove (disk, part);
-			ped_disk_join_freespace (disk);
-			return 0;
-		}
-	}
-
-	check_sanity (disk);
 	return 1;
 }
 
+static int
+_add_partition (PedDisk* disk, PedPartition* part)
+{
+	PedPartition*	walk;
+	PedPartition*	last;
+	PedPartition*	ext_part;
+
+	ext_part = ped_disk_extended_partition (disk);
+
+	last = NULL;
+	walk = (part->type & PED_PARTITION_LOGICAL) ?
+			ext_part->part_list : disk->part_list;
+
+	for (; walk; last = walk, walk = walk->next) {
+		if (walk->geom.start > part->geom.end)
+			break;
+	}
+
+	if (walk) {
+		return _disk_raw_insert_before (disk, walk, part);
+	} else {
+		if (last) {
+			return _disk_raw_insert_after (disk, last, part);
+		} else {
+			if (part->type & PED_PARTITION_LOGICAL)
+				ext_part->part_list = part;
+			else
+				disk->part_list = part;
+		} 
+	}
+
+	return 1;
+}
+
+int
+ped_disk_add_partition (PedDisk* disk, PedPartition* part)
+{
+	PED_ASSERT (disk != NULL, return 0);
+	PED_ASSERT (part != NULL, return 0);
+
+#ifdef VERBOSE
+	printf ("ped_disk_add_partition (dev=\"%s\", start=%d, end=%d,"
+		" type=%x)\n",
+		disk->dev->path, (int) part->geom.start, (int) part->geom.end,
+		(int) part->system);
+#endif
+
+	if (!_check_new_partition_basic_sanity (disk, part))
+		return 0;
+
+	_disk_push_update_mode (disk);
+
+	if (ped_partition_is_active (part)) {
+		if (!ped_disk_enumerate_partition (disk, part))
+			goto error;
+		if (!ped_disk_align_partition (disk, part))
+			goto error;
+	}
+	if (!_check_new_partition (disk, part))
+		goto error;
+	if (!_add_partition (disk, part))
+		goto error;
+
+	_disk_pop_update_mode (disk);
+	if (!_disk_check_sanity (disk))
+		return 0;
+	return 1;
+
+error:
+	_disk_pop_update_mode (disk);
+	return 0;
+}
+
 static int
 ped_disk_delete_all_logical (PedDisk* disk);
 
@@ -1041,20 +1182,17 @@ ped_disk_delete_partition (PedDisk* disk, PedPartition* part)
 
 	is_metadata_part = part->type & PED_PARTITION_METADATA;
 
+	_disk_push_update_mode (disk);
+
 	if (part->type & PED_PARTITION_EXTENDED)
 		ped_disk_delete_all_logical (disk);
+	_disk_raw_remove (disk, part);
+	ped_partition_destroy (part);
+
+	_disk_pop_update_mode (disk);
 
-	part->type = (part->type & PED_PARTITION_LOGICAL)
-			| PED_PARTITION_FREESPACE;
-	part->system = 0;
-	ped_disk_join_freespace (disk);
 	ped_disk_enumerate_partitions (disk);
 
-	if (!is_metadata_part) {
-		if (!ped_disk_alloc_metadata (disk))
-			return 0;
-	}
-
 	return 1;
 }
 
@@ -1062,12 +1200,16 @@ static int
 ped_disk_delete_all_logical (PedDisk* disk)
 {
 	PedPartition*		walk;
+	PedPartition*		next;
+	PedPartition*		ext_part;
 
 	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (ped_disk_extended_partition (disk) != NULL, return 0);
+	ext_part = ped_disk_extended_partition (disk);
+	PED_ASSERT (ext_part != NULL, return 0);
+
+	for (walk = ext_part->part_list; walk; walk = next) {
+		next = walk->next;
 
-	for (walk = ped_disk_extended_partition (disk)->part_list;
-	     walk; walk = walk->next) {
 		if (!(walk->type & PED_PARTITION_FREESPACE)) {
 			if (!ped_disk_delete_partition (disk, walk))
 				return 0;
@@ -1080,95 +1222,62 @@ int
 ped_disk_delete_all (PedDisk* disk)
 {
 	PedPartition*		walk;
+	PedPartition*		next;
 
 	PED_ASSERT (disk != NULL, return 0);
 
-	for (walk = disk->part_list; walk; walk = walk->next) {
-		if (walk->type == PED_PARTITION_EXTENDED) {
-			if (!ped_disk_delete_all_logical (disk))
-				return 0;
-		}
-		if (!(walk->type & PED_PARTITION_FREESPACE)) {
-			if (!ped_disk_delete_partition (disk, walk))
-				return 0;
-		}
+	_disk_push_update_mode (disk);
+
+	for (walk = disk->part_list; walk; walk = next) {
+		next = walk->next;
+
+		if (!ped_disk_delete_partition (disk, walk))
+			return 0;
 	}
-	ped_disk_join_freespace (disk);
+
+	_disk_pop_update_mode (disk);
+
 	return 1;
 }
 
+static PedGeometry*
+_calc_aligned_geom (PedDisk* disk, PedPartition* part,
+		    PedSector start, PedSector end)
+{
+	PedGeometry	old_geom;
+	PedGeometry*	aligned_geom;
+
+	old_geom = part->geom;
+	ped_geometry_set (&part->geom, start, end - start + 1);
+
+	if (ped_partition_is_active (part)) {
+		if (!ped_disk_align_partition (disk, part))
+			goto error;
+	}
+
+	aligned_geom = ped_geometry_duplicate (&part->geom);
+	part->geom = old_geom;
+	return aligned_geom;
+
+error:
+	part->geom = old_geom;
+	return NULL;
+}
+
 static int
-check_new_geometry (PedDisk* disk, PedPartition* part, PedGeometry* old_geom)
-{
-	PedPartition*	left_part;
-	PedPartition*	right_part;
-
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (part != NULL, return 0);
-	PED_ASSERT (old_geom != NULL, return 0);
-
-	left_part = part->prev;
-	right_part = part->next;
-
-	if (part->geom.start > part->geom.end) {
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-			_("Can't move end of the partition behind the start."));
-		return 0;
-	}
-
-/* Assumes that metadata won't get in the way, because align_partition()
- * should prevent this.
- */
-	for (; left_part; left_part = left_part->prev) {
-		if (!left_part->prev
-		    || ped_partition_is_active (left_part->prev))
-			break;
-	}
-	for (; right_part; right_part = right_part->next) {
-		if (!right_part->next
-		    || ped_partition_is_active (right_part->next))
-			break;
-	}
-
-	if (part->geom.start < old_geom->start) {
-		if (!left_part || ped_partition_is_active (left_part))
-			goto used_space; 
-		if (part->geom.start < left_part->geom.start)
-			goto used_space;
-	}
-	if (part->geom.end > old_geom->end) {
-		if (!right_part || ped_partition_is_active (right_part))
-			goto used_space;
-		if (part->geom.end > right_part->geom.end)
-			goto used_space;
-	}	
-	if (part->geom.start > part->geom.end) {
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-			_("Can't move start of partition beyond end."));
-		return 0;
-	}
-	return 1;
-
-used_space:
-	ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-	     _("Can't grow partition onto used space, or outside the disk."));
-	return 0;
-}
-
-int
-check_extended_geometry (PedDisk* disk, PedPartition* ext_part)
+_check_new_extended_geometry (PedDisk* disk, PedGeometry* new_geom)
 {
 	PedPartition*		walk;
+	PedPartition*		ext_part;
 
 	PED_ASSERT (disk != NULL, return 0);
+	ext_part = ped_disk_extended_partition (disk);
 	PED_ASSERT (ext_part != NULL, return 0);
 
 	for (walk = ext_part->part_list; walk; walk = walk->next) {
-		if (!ped_partition_is_active (walk))
-			continue;
-		if (walk->geom.start < ext_part->geom.start
-		    || walk->geom.end > ext_part->geom.end) {
-			ped_exception_throw (PED_EXCEPTION_ERROR,
+		if (!ped_geometry_test_inside (new_geom, &walk->geom)) {
+			ped_exception_throw (
+				PED_EXCEPTION_ERROR,
 				PED_EXCEPTION_CANCEL,
 				_("Can't resize an extended partition so as to "
 				  "exclude a logical partition."));
@@ -1178,145 +1287,119 @@ check_extended_geometry (PedDisk* disk, PedPartition* ext_part)
 	return 1;
 }
 
-static int
-ped_disk_set_partition_start (PedDisk* disk, PedPartition* part,
-			      PedSector start)
-{
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (part != NULL, return 0);
-
-	if (start > part->geom.start) {
-		if (!ped_disk_split_left (disk, part, start))
-			return 0;
-		ped_disk_join_freespace (disk);
-	}
-	if (start < part->geom.start) {
-		if (start > part->prev->geom.start) {
-			part->prev->geom.end = start - 1;
-			part->prev->geom.length = part->prev->geom.end
-						  - part->prev->geom.start + 1;
-		} else {
-			disk_raw_remove (disk, part->prev);
-		}
-		part->geom.start = start;
-		part->geom.length = part->geom.end - part->geom.start + 1;
-	}
-	return 1;
-}
-
-static int
-ped_disk_set_partition_end (PedDisk* disk, PedPartition* part, PedSector end)
-{
-	PED_ASSERT (disk != NULL, return 0);
-	PED_ASSERT (part != NULL, return 0);
-
-	if (end < part->geom.end) {
-		if (!ped_disk_split_right (disk, part, end))
-			return 0;
-		ped_disk_join_freespace (disk);
-	}
-	if (end > part->geom.end) {
-		if (end < part->next->geom.end) {
-			part->next->geom.start = end + 1;
-			part->next->geom.length = part->next->geom.end
-						  - part->next->geom.start + 1;
-		} else {
-			disk_raw_remove (disk, part->next);
-		}
-		part->geom.end = end;
-		part->geom.length = part->geom.end - part->geom.start + 1;
-	}
-	return 1;
-}
-
-/* Arrghhh.  This code is a mess.
+/* FIXME: do metadata collion handling properly! (we're depending on being
+ * in update mode, at the moment, which is yucky)
  */
+static int
+_check_changed_geometry (PedDisk* disk, PedPartition* part,
+			 PedGeometry* new_geom)
+{
+	PedPartition*	ext_part;
+
+	PED_ASSERT (disk != NULL, return 0);
+	PED_ASSERT (part != NULL, return 0);
+	PED_ASSERT (new_geom != NULL, return 0);
+
+	if (new_geom->start > new_geom->end) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("Can't move the end of the partition behind the "
+			  "start."));
+		goto error;
+	}
+
+	if (part->type & PED_PARTITION_LOGICAL) {
+		ext_part = ped_disk_extended_partition (disk);
+		if (!ped_geometry_test_inside (&ext_part->geom, new_geom)) {
+			ped_exception_throw (
+				PED_EXCEPTION_ERROR,
+				PED_EXCEPTION_CANCEL,
+				_("Can't grow a logical partition outside "
+				  "the extended partition."));
+			return 0;
+		}
+	}
+
+	if (part->type & PED_PARTITION_EXTENDED) {
+		if (!_check_new_extended_geometry (disk, new_geom))
+			goto error;
+	}
+
+	if (!ped_disk_check_overlap_new_geom (disk, part, new_geom)) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("Can't grow partition onto used space, or outside "
+			  "the disk."));
+		goto error;
+	}
+
+	return 1;
+
+error:
+	return 0;
+}
+
 int
 ped_disk_set_partition_geom (PedDisk* disk, PedPartition* part,
 			     PedSector start, PedSector end)
 {
 	PedPartition*		free_space;
 	PedPartition*		ext_free_space;
-	PedGeometry		old_geom;
-	PedGeometry		new_geom;
+	PedGeometry*		new_geom;
 
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (part != NULL, return 0);
 
-	old_geom = part->geom;
-	part->geom.start = start;
-	part->geom.end = end;
-	part->geom.length = end - start + 1;
+	new_geom = _calc_aligned_geom (disk, part, start, end);
+	if (!new_geom)
+		goto error;
 
-	if (ped_partition_is_active (part)) {
-		ped_disk_remove_metadata (disk);
-		if (!ped_disk_align_partition (disk, part))
-			goto error_restore_geom;
-	}
-	if (!check_new_geometry (disk, part, &old_geom))
-		goto error_restore_geom;
-	if (part->type == PED_PARTITION_EXTENDED) {
-		if (!check_extended_geometry (disk, part))
-			goto error_restore_geom;
-	}
+	_disk_push_update_mode (disk);
 
-/* now that we know we can do it, now we need to do it properly */
-	new_geom = part->geom;
-	part->geom = old_geom;
+	if (!_check_changed_geometry (disk, part, new_geom))
+		goto error;
 
-	if (new_geom.start < part->geom.end) {
-		if (!ped_disk_set_partition_start (disk, part, new_geom.start))
-			goto error_weird;
-		if (!ped_disk_set_partition_end (disk, part, new_geom.end))
-			goto error_weird;
-	} else {
-		if (!ped_disk_set_partition_end (disk, part, new_geom.end))
-			goto error_weird;
-		if (!ped_disk_set_partition_start (disk, part, new_geom.start))
-			goto error_weird;
-	}
+	part->geom = *new_geom;
+	ped_geometry_destroy (new_geom);
 
-	if (part->type == PED_PARTITION_EXTENDED) {
-		ext_free_space = part->part_list;	/* XXX */
-		ped_disk_set_partition_geom (disk, ext_free_space,
-					     start, ext_free_space->geom.end);
-	}
-
-	if (ped_partition_is_active (part)) {
-		if (!ped_disk_alloc_metadata (disk))
-			goto error;
-	}
+	_disk_pop_update_mode (disk);
 
 	return 1;
 
-error_weird:
-	ped_exception_throw (PED_EXCEPTION_BUG, PED_EXCEPTION_CANCEL,
-			     _("Weird bug in ped_disk_set_partition_geom()"));
-error_restore_geom:
-	part->geom = old_geom;
-error_restore_meta:
-	if (ped_partition_is_active (part))
-		ped_disk_alloc_metadata (disk);
 error:
+	_disk_pop_update_mode (disk);
 	return 0;
 }
 
 int
 ped_disk_maximize_partition (PedDisk* disk, PedPartition* part)
 {
+	PedGeometry	old_geom;
+
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (part != NULL, return 0);
 
+	old_geom = part->geom;
+
 	if (part->prev && part->prev->type & PED_PARTITION_FREESPACE) {
-		part->geom.start = part->prev->geom.start;
-		disk_raw_remove (disk, part->prev);
+		if (!ped_disk_set_partition_geom (
+				disk, part, part->prev->geom.start,
+				part->geom.end))
+			goto error;
 	}
 	if (part->next && part->next->type & PED_PARTITION_FREESPACE) {
-		part->geom.end = part->next->geom.end;
-		disk_raw_remove (disk, part->next);
+		if (!ped_disk_set_partition_geom (
+				disk, part, part->geom.start,
+				part->next->geom.end))
+			goto error;
 	}
-	part->geom.length = part->geom.end - part->geom.start + 1;
 	return 1;
+
+error:
+	ped_disk_set_partition_geom (disk, part, old_geom.start, old_geom.end);
+	return 0;
 }
 
 /* Reduces the size of the extended partition to wrap the logical partitions.
@@ -1325,35 +1408,27 @@ ped_disk_maximize_partition (PedDisk* disk, PedPartition* part)
 int
 ped_disk_minimize_extended_partition (PedDisk* disk)
 {
-	PedPartition*		part;
+	PedPartition*		first_logical;
+	PedPartition*		last_logical;
 	PedPartition*		walk;
-	PedSector		cut_point;
+	PedPartition*		ext_part;
 
 	PED_ASSERT (disk != NULL, return 0);
 
-	part = ped_disk_extended_partition (disk);
-	if (!part)
+	ext_part = ped_disk_extended_partition (disk);
+	if (!ext_part)
 		return 1;
 
-	if (part->part_list->type & PED_PARTITION_FREESPACE) {
-		cut_point = part->part_list->geom.end + 1;
-		disk_raw_remove (disk, part->part_list);
-		ped_disk_split_left (disk, part, cut_point);
-		if (!part->part_list) {
-			ped_disk_delete_partition (disk, part);
-			return 1;
-		}
-	}
+	first_logical = ext_part->part_list;
+	if (!first_logical)
+		return ped_disk_delete_partition (disk, ext_part);
 
-	for (walk = part->part_list; walk->next; walk = walk->next);
+	for (walk = first_logical; walk->next; walk = walk->next);
+	last_logical = walk;
 
-	if (walk->type & PED_PARTITION_FREESPACE) {
-		cut_point = walk->geom.start - 1;
-		disk_raw_remove (disk, walk);
-		ped_disk_split_right (disk, walk, cut_point);
-	}
-
-	return 1;
+	return ped_disk_set_partition_geom (disk, ext_part,
+					    first_logical->geom.start,
+					    last_logical->geom.end);
 }
 
 char*
