@@ -54,7 +54,8 @@ static char MBR_BOOT_CODE[] = {
 	0x00, 0xeb, 0xfe
 };
 
-#define MSDOS_MAGIC	0xAA55
+#define MSDOS_MAGIC		0xAA55
+#define PARTITION_MAGIC_MAGIC	0xf6f6
 
 /* ripped from Linux source */
 struct _DosRawPartition {
@@ -88,7 +89,7 @@ static int msdos_write (PedDisk* disk);
 static int msdos_align_partition (PedDisk* disk, PedPartition* part);
 static int msdos_enumerate_partition (PedDisk* disk, PedPartition* part);
 static int msdos_alloc_metadata (PedDisk* disk);
-static char msdos_get_extended_system ();
+static int msdos_get_extended_system ();
 
 static PedDiskOps msdos_disk_ops = {
 	probe:			msdos_probe,
@@ -100,8 +101,8 @@ static PedDiskOps msdos_disk_ops = {
 	
 	align_partition:	msdos_align_partition,
 	enumerate_partition:	msdos_enumerate_partition,
-	alloc_metadata:		msdos_alloc_metadata,
-	get_extended_system:	msdos_get_extended_system
+	get_extended_system:	msdos_get_extended_system,
+	alloc_metadata:		msdos_alloc_metadata
 };
 
 static PedDiskType msdos_disk_type = {
@@ -154,13 +155,10 @@ msdos_open (PedDevice* dev)
 		goto error;
 
 	ped_device_open ((PedDevice*) dev);
-	disk = (PedDisk*) ped_malloc (sizeof (PedDisk));
+	disk = ped_disk_alloc (dev, &msdos_disk_type);
 	if (!disk)
 		goto error;
 
-	disk->dev = (PedDevice*) dev;
-	disk->type = &msdos_disk_type;
-
 	disk->part_list = ped_partition_new (
 		disk, PED_PARTITION_FREESPACE, NULL,
 		0, dev->length);
@@ -186,14 +184,27 @@ error:
 static void
 query_write_boot_code (PedDevice* dev, DosRawTable* table)
 {
-	if (ped_exception_throw (PED_EXCEPTION_INFORMATION,
+	PedExceptionOption	ex_status;
+
+	ex_status = ped_exception_throw (PED_EXCEPTION_INFORMATION,
 		PED_EXCEPTION_YES_NO,
 		_("There is no boot code on this disk.  Would you like "
-		  "to use GNU parted's boot loader?"))
-			!= PED_EXCEPTION_NO) {
-		memset (table->boot_code, 0, 512);
-		memcpy (table->boot_code, MBR_BOOT_CODE,
-			sizeof (MBR_BOOT_CODE));
+		  "to use GNU parted's boot loader?"));
+
+	switch (ex_status) {
+		case PED_EXCEPTION_NO:
+			return;
+
+		case PED_EXCEPTION_UNHANDLED:
+			ped_exception_catch ();
+		case PED_EXCEPTION_YES:
+			memset (table->boot_code, 0, 512);
+			memcpy (table->boot_code, MBR_BOOT_CODE,
+				sizeof (MBR_BOOT_CODE));
+/*
+			table->boot_code [MBR_BOOT_CODE_DISK_OFFSET]
+				= 0x80 + dev->host * 2 + dev->did;
+*/
 	}
 }
 
@@ -237,9 +248,9 @@ raw_part_is_extended (DosRawPartition* raw_part)
 	PED_ASSERT (raw_part != NULL, return 0);
 
 	switch (raw_part->type) {
-	case PED_DOS_EXTENDED:
-	case PED_LINUX_EXTENDED:
-	case PED_WIN95_EXTENDED:
+	case 0x05:
+	case 0x0f:
+	case 0x85:
 		return 1;
 
 	default:
@@ -253,12 +264,12 @@ raw_part_is_hidden (DosRawPartition* raw_part)
 	PED_ASSERT (raw_part != NULL, return 0);
 
 	switch (raw_part->type) {
-	case PED_DOS_HIDDEN_FAT12:
-	case PED_DOS_HIDDEN_FAT16:
-	case PED_DOS_HIDDEN_FAT16_BIG:
-	case PED_WIN95_HIDDEN_FAT32:
-	case PED_WIN95_HIDDEN_FAT32_LBA:
-	case PED_WIN95_HIDDEN_FAT16_LBA:
+	case 0x11:
+	case 0x14:
+	case 0x16:
+	case 0x1b:
+	case 0x1c:
+	case 0x1e:
 	return 1;
 
 	default:
@@ -334,7 +345,8 @@ legacy_end (PedDisk* disk, const DosRawPartition* raw_part)
 }
 
 static PedSector
-linear_start (PedDisk* disk, const DosRawPartition* raw_part, PedSector offset)
+linear_start (const PedDisk* disk, const DosRawPartition* raw_part,
+	      PedSector offset)
 {
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (raw_part != NULL, return 0);
@@ -345,13 +357,14 @@ linear_start (PedDisk* disk, const DosRawPartition* raw_part, PedSector offset)
 }
 
 static PedSector
-linear_end (PedDisk* disk, const DosRawPartition* raw_part, PedSector offset)
+linear_end (const PedDisk* disk, const DosRawPartition* raw_part,
+	    PedSector offset)
 {
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (raw_part != NULL, return 0);
 
 	return linear_start (disk, raw_part, offset)
-	       + PED_LE32_TO_CPU (raw_part->length - 1)
+	       + (PED_LE32_TO_CPU (raw_part->length) - 1)
 	       	 	* (disk->dev->sector_size / 512);
 }
 
@@ -442,7 +455,7 @@ probe_acceptable_geom (PedDisk* disk, const DosRawPartition* raw_part,
 	PedSector	start, end;
 	PedSector	heads, sectors;
 	PedSector	tmp_heads, tmp_sectors;
-	PedSector	h_factor, s_factor;	/* factor of h&s to pull out */
+	PedSector	h_factor, s_factor;	/* factor of h&s to remove */
 
 	start = linear_start (disk, raw_part, offset);
 	end = linear_end (disk, raw_part, offset);
@@ -485,7 +498,8 @@ static int
 process_inconsistent_table (PedDisk* disk, const DosRawPartition* raw_part,
 			    PedSector offset)
 {
-	int old_c, old_h, old_s;
+	PedExceptionOption	ex_status;
+	int			old_c, old_h, old_s;
 
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (raw_part != NULL, return 0);
@@ -495,7 +509,7 @@ process_inconsistent_table (PedDisk* disk, const DosRawPartition* raw_part,
 	old_s = disk->dev->sectors;
 
 	if (probe_real_geom (disk, raw_part, offset)) {
-		if (ped_exception_throw (
+		ex_status = ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_IGNORE_CANCEL,
 			_("The partition table on %s is inconsistent.  "
@@ -522,29 +536,42 @@ process_inconsistent_table (PedDisk* disk, const DosRawPartition* raw_part,
 			disk->dev->path + strlen ("/dev/"),
 			(int) disk->dev->cylinders,
 			(int) disk->dev->heads,
-			(int) disk->dev->sectors)
-				!= PED_EXCEPTION_IGNORE) {
-			disk->dev->cylinders = old_c;
-			disk->dev->heads = old_h;
-			disk->dev->sectors = old_s;
-			return 0;
+			(int) disk->dev->sectors);
+
+		switch (ex_status) {
+			case PED_EXCEPTION_CANCEL:
+				disk->dev->cylinders = old_c;
+				disk->dev->heads = old_h;
+				disk->dev->sectors = old_s;
+				return 0;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_IGNORE:
+				return 1;
 		}
 	} else {
-		if (ped_exception_throw (
-			PED_EXCEPTION_ERROR,
+		ex_status = ped_exception_throw (
+			PED_EXCEPTION_WARNING,
 			PED_EXCEPTION_IGNORE_CANCEL,
 			_("The partition table on %s is inconsistent.  "
 			"There are many reasons why this might be the case.  "
 			"Often, the reason is that Linux detected the BIOS "
 			"geometry incorrectly.  However, this does not "
 			"appear to be the case here."),
-			disk->dev->path)
-				!= PED_EXCEPTION_IGNORE)
-			return 0;
+			disk->dev->path);
 
-		probe_acceptable_geom (disk, raw_part, offset);
+		switch (ex_status) {
+			case PED_EXCEPTION_CANCEL:
+				return 0;
+
+			case PED_EXCEPTION_UNHANDLED:
+				ped_exception_catch ();
+			case PED_EXCEPTION_IGNORE:
+				return probe_acceptable_geom (disk, raw_part,
+							      offset);
+		}
 	}
-	return 1;
 }
 
 static int
@@ -584,19 +611,25 @@ read_table (PedDisk* disk, PedSector sector, int is_extended_table)
 
 	if (!ped_device_read (disk->dev, (void*) &table, sector, 1))
 		goto error;
+
+	/* weird: empty extended partitions are filled with 0xf6 by PM */
+	if (is_extended_table
+	    && PED_LE16_TO_CPU (table.magic) == PARTITION_MAGIC_MAGIC)
+		return 1;
+
 	if (PED_LE16_TO_CPU (table.magic) != MSDOS_MAGIC) {
 		if (ped_exception_throw (
 			PED_EXCEPTION_ERROR, PED_EXCEPTION_IGNORE_CANCEL,
 			_("Invalid partition table on %s - wrong signature %x"),
 			disk->dev->path,
-			PED_LE16_TO_CPU (table.magic)
-				!= PED_EXCEPTION_IGNORE))
+			PED_LE16_TO_CPU (table.magic))
+				!= PED_EXCEPTION_IGNORE)
 			goto error;
 	}
 
 	for (i = 0; i < 4; i++) {
 		raw_part = &table.partitions [i];
-		if (raw_part->type == PED_FREE_SPACE || !raw_part->length)
+		if (raw_part->type == 0x0 || !raw_part->length)
 			continue;
 
 		if (is_extended_table) {
@@ -701,7 +734,7 @@ fill_raw_part (DosRawPartition* raw_part, PedPartition* part, PedSector offset)
 
 	raw_part->boot_ind = 0x80 * part->bootable;
 	raw_part->type = part->system;
-	raw_part->start = PED_CPU_TO_LE32 (part->geom.start - offset
+	raw_part->start = PED_CPU_TO_LE32 ((part->geom.start - offset)
 				/ (part->geom.disk->dev->sector_size / 512));
 	raw_part->length = PED_CPU_TO_LE32 (part->geom.length
 				/ (part->geom.disk->dev->sector_size / 512));
@@ -725,8 +758,8 @@ fill_ext_raw_part_geom (DosRawPartition* raw_part, PedGeometry* geom,
 	PED_ASSERT (geom->disk->dev != NULL, return 0);
 
 	raw_part->boot_ind = 0;
-	raw_part->type = PED_DOS_EXTENDED;
-	raw_part->start = PED_CPU_TO_LE32 (geom->start - offset
+	raw_part->type = 0x5;
+	raw_part->start = PED_CPU_TO_LE32 ((geom->start - offset)
 				/ (geom->disk->dev->sector_size / 512));
 	raw_part->length = PED_CPU_TO_LE32 (geom->length
 				/ (geom->disk->dev->sector_size / 512));
@@ -862,10 +895,18 @@ round_up_to (PedSector sector, PedSector grain_size)
 		return sector;
 }
 
+static PedSector
+round_to_nearest (PedSector sector, PedSector grain_size)
+{
+	if (sector % grain_size > grain_size/2)
+		return round_up_to (sector, grain_size);
+	else
+		return round_down_to (sector, grain_size);
+}
+
 static int
 msdos_align_partition (PedDisk* disk, PedPartition* part)
 {
-	PedPartition*		current;
 	PedPartition*		ext_part;
 	PedDevice*		dev;
 	PedSector		cylinder_size;
@@ -878,17 +919,16 @@ msdos_align_partition (PedDisk* disk, PedPartition* part)
 	dev = disk->dev;
 	cylinder_size = dev->sectors * dev->heads;
 
-	current = ped_disk_get_partition_by_sector (disk, part->geom.start);
-	if (!current) {
-		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-			_("Can't align geometry outside of device."));
-		return 0;
-	}
-
-	part->geom.start = round_down_to (part->geom.start, cylinder_size);
+	part->geom.start = round_to_nearest (part->geom.start, cylinder_size);
 	if (part->type & PED_PARTITION_LOGICAL) {
 		PED_ASSERT (ext_part != NULL, return 0);
 
+		/* if it's a logical partition outside an extended partition,
+		 * back out immediately!  Let the support code deal with it...
+		 */
+		if (part->geom.start <= ext_part->geom.start - cylinder_size)
+			return 1;
+
 		if (part->geom.start <= ext_part->geom.start + dev->sectors) {
 			part->geom.start = ext_part->geom.start + dev->sectors;
 			if (part->num > 5)
@@ -899,7 +939,7 @@ msdos_align_partition (PedDisk* disk, PedPartition* part)
 	} else if (part->geom.start == 0) {
 		part->geom.start += dev->sectors;
 	}
-	part->geom.end = round_up_to (part->geom.end, cylinder_size) - 1;
+	part->geom.end = round_to_nearest (part->geom.end, cylinder_size) - 1;
 	part->geom.length = part->geom.end - part->geom.start + 1;
 	return 1;
 }
@@ -1021,9 +1061,9 @@ msdos_enumerate_partition (PedDisk* disk, PedPartition* part)
 	return 1;
 }
 
-static char
+static int
 msdos_get_extended_system ()
 {
-	return PED_DOS_EXTENDED;
+	return 0x5;
 }
 
