@@ -35,6 +35,7 @@
 
 #include <stdio.h>
 #include <stdarg.h>
+#include <stdlib.h>
 
 int				ped_exception = 0;
 
@@ -44,7 +45,7 @@ static PedExceptionHandler*	ex_handler = default_handler;
 static PedException*		ex = NULL;
 static int			ex_fetch_count = 0;
 
-static char*	class_strings [] = {
+static char*	type_strings [] = {
 	N_("Information"),
 	N_("Warning"),
 	N_("Error"),
@@ -64,9 +65,9 @@ static char*	option_strings [] = {
 };
 
 char*
-ped_exception_get_class_string (PedExceptionClass ex_class)
+ped_exception_get_type_string (PedExceptionType ex_type)
 {
-	return class_strings [ex_class - 1];
+	return type_strings [ex_type - 1];
 }
 
 static int
@@ -86,27 +87,24 @@ ped_exception_get_option_string (PedExceptionOption ex_opt)
 static PedExceptionOption
 default_handler (PedException* ex)
 {
-	if (ex->class == PED_EXCEPTION_BUG)
+	if (ex->type == PED_EXCEPTION_BUG)
 		printf (_("A bug has been detected in GNU parted.  "
-			"Please email a bug report to\n"
+			"Please email a bug report to "
 			"bug-parted@gnu.org containing the version (%s) "
-			"and the following message:\n"),
+			"and the following message:"),
 			VERSION);
 	else
-		printf ("%s: ", ped_exception_get_class_string (ex->class));
+		printf ("%s: ", ped_exception_get_type_string (ex->type));
 	printf ("%s\n", ex->message);
 
-	switch (ex->class) {
-	case PED_EXCEPTION_INFORMATION:
-	case PED_EXCEPTION_WARNING:
-		return PED_EXCEPTION_OK;
+	switch (ex->options) {
+		case PED_EXCEPTION_OK:
+		case PED_EXCEPTION_CANCEL:
+		case PED_EXCEPTION_IGNORE:
+			return ex->options;
 
-	case PED_EXCEPTION_ERROR:
-	case PED_EXCEPTION_FATAL:
-	case PED_EXCEPTION_BUG:
-	case PED_EXCEPTION_NO_FEATURE:
-	default:
-		return PED_EXCEPTION_CANCEL;
+		default:
+			return PED_EXCEPTION_UNHANDLED;
 	}
 }
 
@@ -145,7 +143,7 @@ do_throw ()
 }
 
 PedExceptionOption
-ped_exception_throw (PedExceptionClass ex_class,
+ped_exception_throw (PedExceptionType ex_type,
 		     PedExceptionOption ex_opts, const char* message, ...)
 {
 	va_list		arg_list;
@@ -158,7 +156,7 @@ ped_exception_throw (PedExceptionClass ex_class,
 	if (!ex->message)
 		goto no_memory;
 
-	ex->class = ex_class;
+	ex->type = ex_type;
 	ex->options = ex_opts;
 
 	va_start (arg_list, message);
