@@ -725,9 +725,6 @@ struct ext2_fs *ext2_open(struct ext2_dev_handle *handle, int state)
 		== NULL)
 		goto error;
 
-	if ((fs->gd = (struct ext2_group_desc *) ped_malloc(ext2_max_groups * sizeof(struct ext2_group_desc))) == NULL)
-		goto error_free_fs;
-
 	handle->ops->set_blocksize(handle->cookie, 10);
 
 	if (!handle->ops->read(handle->cookie, &fs->sb, 1, 1)
@@ -736,9 +733,10 @@ struct ext2_fs *ext2_open(struct ext2_dev_handle *handle, int state)
 		ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
 			_("Invalid superblock.  Are you sure this is an ext2 "
 			  "filesystem?"));
-		goto error_free_gd;
+		goto error_free_fs;
 	}
 
+	
 	fs->opt_debug = 1;
 	fs->opt_safe = 1;
 	fs->opt_verbose = 0;
@@ -790,6 +788,11 @@ struct ext2_fs *ext2_open(struct ext2_dev_handle *handle, int state)
 	if (fs->sb.s_feature_ro_compat & EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)
 		fs->sparse = 1;
 
+	fs->gd = ped_malloc (fs->numgroups * sizeof (struct ext2_group_desc)
+			     	+ fs->blocksize);
+	if (!fs->gd)
+		goto error_free_fs;
+
 	ext2_read_blocks(fs, fs->gd, fs->sb.s_first_data_block + 1, fs->gdblocks);
 
 	if (!ext2_determine_itoffset(fs))
