@@ -252,5 +252,5 @@ _ext2_get_system (const PedFileSystem* fs, const PedPartition* part,
 				     _("Ext2 partitions can't be hidden."));
 		return 0;
 	}
-	return PED_LINUX;
+	return 0x83;
 }
