@@ -40,22 +40,17 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#define BUFFER_SIZE  64		/* buffer size in clusters */
+#define BUFFER_SIZE  1024	/* buffer size in sectors (512 bytes) */
 
-typedef unsigned int            FatCluster;
-typedef unsigned long long	FatLogicalSector;
-
-typedef __u8		FatFlag;
-#define	FAT_FLAG_FREE		0
-#define FAT_FLAG_FILE		1
-#define	FAT_FLAG_DIRECTORY	2
+typedef __s32			FatCluster;
+typedef __s32			FatFragment;
 
 enum _FatType {
 	FAT_TYPE_FAT12,
 	FAT_TYPE_FAT16,
 	FAT_TYPE_FAT32
 };
-typedef enum _FatType FatType;
+typedef enum _FatType		FatType;
 
 typedef struct _FatSpecific	FatSpecific;
 typedef struct _FatDirEntry	FatDirEntry;
@@ -105,7 +100,7 @@ struct _FatSpecific {
 	PedSector	info_sector_offset;     /* FAT32 only */
 	PedSector	fat_offset;
 	PedSector	root_dir_offset;	/* non-FAT32 */
-	PedSector	cluster_offset;		/* NOTE: is -2*cluster size */
+	PedSector	cluster_offset;
 	PedSector	boot_sector_backup_offset;
 
 	FatCluster	root_cluster;           /* FAT32 only */
@@ -114,11 +109,16 @@ struct _FatSpecific {
 	FatCluster	total_dir_clusters;
 
 	FatTable*	fat;
+	FatClusterInfo*	cluster_info;
 
+	PedSector	buffer_sectors;
 	char*		buffer;
-	FatCluster	buffer_size;            /* buffer size in clusters */
 
-	FatFlag*	fat_flag_map;
+	int		frag_size;
+	PedSector	frag_sectors;
+	FatFragment	frag_count;
+	FatFragment	buffer_frags;
+	FatFragment	cluster_frags;
 };
 
 #define FAT_SPECIFIC(fs)	((FatSpecific*) fs->type_specific)
@@ -151,15 +151,20 @@ extern char fat_get_system (const PedFileSystem* fs, const PedPartition* part,
 			    PedDiskType* disk_type);
 extern void fat_print (const PedFileSystem* fs);
 
-extern PedFileSystem* fat_file_system_alloc (const PedGeometry* geom);
-extern void fat_file_system_free (PedFileSystem* fs);
-extern int fat_file_system_alloc_buffers (PedFileSystem* fs);
-extern void fat_file_system_free_buffers (PedFileSystem* fs);
+extern PedFileSystem* fat_alloc (const PedGeometry* geom);
+extern void fat_free (PedFileSystem* fs);
+extern int fat_alloc_buffers (PedFileSystem* fs);
+extern void fat_free_buffers (PedFileSystem* fs);
 
-inline static unsigned int
-div_round_up (unsigned int a, unsigned int b)
+extern int fat_set_frag_sectors (PedFileSystem* fs, PedSector frag_sectors);
+
+#define MIN(a,b)	((a<b)? a : b)
+#define MAX(a,b)	(a>b) ? a : b;
+
+inline static PedSector
+div_round_up (PedSector a, PedSector b)
 {
-	int	result;
+	PedSector	result;
 
 	result = a / b;
 	if (a % b) result++;
