@@ -33,14 +33,6 @@
 #include <stdarg.h>
 #include <string.h>
 
-#ifdef USE_NEWT
-#include <newt.h>
-#endif
-
-#define MIN(a, b) (a<b) ? a : b;
-#define MAX(a, b) (a>b) ? a : b;
-#define BUFFER_SIZE 64
-
 /*
     Recursively duplicates a directory tree, modified for the new file
     locations.
@@ -50,15 +42,18 @@ fat_construct_directory (FatOpContext* ctx, FatTraverseInfo* trav_info)
 {
 	FatTraverseInfo*	sub_dir_info;
 	FatDirEntry*		dir_entry;
+	FatCluster		old_first_cluster;
 
 	while ( (dir_entry = fat_traverse_next_dir_entry (trav_info)) ) {
 		if (!fat_dir_entry_has_first_cluster (dir_entry, ctx->old_fs))
 			continue;
 
 		fat_traverse_mark_dirty (trav_info);
+
+		old_first_cluster = fat_dir_entry_get_first_cluster (dir_entry,
+						ctx->old_fs);
 		fat_dir_entry_set_first_cluster (dir_entry, ctx->new_fs,
-			ctx->remap [fat_dir_entry_get_first_cluster
-					(dir_entry, ctx->old_fs)]);
+			fat_op_context_map_cluster (ctx, old_first_cluster));
 
 		if (fat_dir_entry_is_directory (dir_entry)
 				&& dir_entry->name [0] != '.') {
@@ -114,7 +109,8 @@ fat_construct_legacy_root (FatOpContext* ctx)
     Constructs the new directory tree for new (FAT32) filesystems.
 */
 static int
-fat_construct_root (FatOpContext* ctx) {
+fat_construct_root (FatOpContext* ctx)
+{
 	FatSpecific*		new_fs_info = FAT_SPECIFIC (ctx->new_fs);
 	FatTraverseInfo*	trav_info;
 
@@ -139,13 +135,10 @@ fat_convert_directory (FatOpContext* ctx, FatTraverseInfo* old_trav,
 	FatTraverseInfo*	sub_new_dir_trav;
 	FatDirEntry*		new_dir_entry;
 	FatDirEntry*		old_dir_entry;
+	FatCluster		old_first_cluster;
 
 	while ( (old_dir_entry = fat_traverse_next_dir_entry (old_trav)) ) {
 		new_dir_entry = fat_traverse_next_dir_entry (new_trav);
-
-		if (!fat_dir_entry_is_active (old_dir_entry))
-			continue;
-
 		if (!new_dir_entry) {
 			return ped_exception_throw (PED_EXCEPTION_ERROR,
 				PED_EXCEPTION_IGNORE_CANCEL,
@@ -155,6 +148,9 @@ fat_convert_directory (FatOpContext* ctx, FatTraverseInfo* old_trav,
 					== PED_EXCEPTION_IGNORE;
 		}
 
+		if (!fat_dir_entry_is_active (old_dir_entry))
+			continue;
+
 		*new_dir_entry = *old_dir_entry;
 		fat_traverse_mark_dirty (new_trav);
 
@@ -162,9 +158,10 @@ fat_convert_directory (FatOpContext* ctx, FatTraverseInfo* old_trav,
 						      ctx->old_fs))
 			continue;
 
+		old_first_cluster = fat_dir_entry_get_first_cluster (
+						old_dir_entry, ctx->old_fs);
 		fat_dir_entry_set_first_cluster (new_dir_entry, ctx->new_fs,
-			ctx->remap [fat_dir_entry_get_first_cluster
-					(old_dir_entry, ctx->old_fs)]);
+			fat_op_context_map_cluster (ctx, old_first_cluster));
 
 		if (fat_dir_entry_is_directory (old_dir_entry)
 				&& old_dir_entry->name [0] != '.') {
@@ -297,35 +294,87 @@ fat_construct_dir_tree (FatOpContext* ctx)
 	return 0;
 }
 
+static FatFragment
+_get_next_old_frag (FatOpContext* ctx, FatFragment frag)
+{
+	FatSpecific*	old_fs_info = FAT_SPECIFIC (ctx->old_fs);
+	FatCluster	cluster;
+	FatCluster	next_cluster;
+
+	if ((frag + 1) % old_fs_info->cluster_frags != 0) {
+		if (fat_is_fragment_active (ctx->old_fs, frag + 1))
+			return frag + 1;
+		else
+			return -1;
+	} else {
+		cluster = fat_frag_to_cluster (ctx->old_fs, frag);
+		next_cluster = fat_table_get (old_fs_info->fat, cluster);
+
+		if (fat_table_is_eof (old_fs_info->fat, next_cluster))
+			return -1;
+		else
+			return fat_cluster_to_frag (ctx->old_fs, next_cluster);
+	}
+}
+
 /*
     constructs the new fat for the resized filesystem.
 */
 static int
-fat_construct_new_fat (FatOpContext* ctx) {
+fat_construct_new_fat (FatOpContext* ctx)
+{
 	FatSpecific*	old_fs_info = FAT_SPECIFIC (ctx->old_fs);
 	FatSpecific*	new_fs_info = FAT_SPECIFIC (ctx->new_fs);
-	FatCluster	old_cluster;
+	FatFragment	old_frag;
 	FatCluster	new_cluster;
-	FatCluster	old_next_cluster;
+	FatFragment	new_frag;
+	FatFragment	old_next_frag;
+	FatFragment	new_next_frag;
+	FatCluster	new_next_cluster;
+	FatClusterFlag	flag;
 	int		i;
 
 	fat_table_clear (new_fs_info->fat);
-	new_fs_info->fat->cluster_count = new_fs_info->cluster_count;
+	if (!fat_table_set_cluster_count (new_fs_info->fat,
+					  new_fs_info->cluster_count))
+		return 0;
 
-	for (old_cluster=2; old_cluster < old_fs_info->cluster_count + 2;
-	     old_cluster++) {
-		if (old_fs_info->fat_flag_map [old_cluster] == FAT_FLAG_FREE)
+	for (old_frag = 0; old_frag < old_fs_info->frag_count; old_frag++) {
+		flag = fat_get_fragment_flag (ctx->old_fs, old_frag);
+		if (flag == FAT_FLAG_FREE)
 			continue;
-			
-		old_next_cluster = fat_table_get (old_fs_info->fat, old_cluster);
-		new_cluster = ctx->remap [old_cluster];
-		if (fat_table_is_eof (old_fs_info->fat, old_next_cluster))
+		if (flag == FAT_FLAG_BAD) {
+			new_frag = fat_op_context_map_static_fragment (
+						ctx, old_frag);
+			if (new_frag == -1)
+				continue;
+			new_cluster = fat_frag_to_cluster (ctx->new_fs,
+							   new_frag);
+			fat_table_set_bad (new_fs_info->fat, new_cluster);
+			continue;
+		}
+
+		new_frag = fat_op_context_map_fragment (ctx, old_frag);
+		new_cluster = fat_frag_to_cluster (ctx->new_fs, new_frag);
+
+		old_next_frag = _get_next_old_frag (ctx, old_frag);
+		if (old_next_frag == -1) {
 			fat_table_set_eof (new_fs_info->fat, new_cluster);
-		else
-			fat_table_set (new_fs_info->fat, new_cluster,
-				       ctx->remap [old_next_cluster]);
+			continue;
+		}
+
+		new_next_frag = fat_op_context_map_fragment (ctx,
+							     old_next_frag);
+		PED_ASSERT (new_next_frag != -1, return 0);
+
+		new_next_cluster = fat_frag_to_cluster (ctx->new_fs,
+							new_next_frag);
+		PED_ASSERT (new_next_cluster != new_cluster, return 0);
+
+		fat_table_set (new_fs_info->fat, new_cluster, new_next_cluster);
 	}
 
+#if 0
 #ifdef VERBOSE
 	for (old_cluster=2; old_cluster < old_fs_info->cluster_count+2;
 	     old_cluster++) {
@@ -340,11 +389,13 @@ fat_construct_new_fat (FatOpContext* ctx) {
 				       ctx->remap [old_cluster]));
 	}
 #endif /* VERBOSE */
+#endif
 
 	if (old_fs_info->fat_type == FAT_TYPE_FAT32
 	    && new_fs_info->fat_type == FAT_TYPE_FAT32) {
 		new_fs_info->root_cluster
-			= ctx->remap [old_fs_info->root_cluster];
+			= fat_op_context_map_cluster (ctx,
+					old_fs_info->root_cluster);
 	}
 
 	if (old_fs_info->fat_type == FAT_TYPE_FAT16
@@ -360,6 +411,101 @@ fat_construct_new_fat (FatOpContext* ctx) {
 	return 1;
 }
 
+/*  For resize operations: determine if the file system must be FAT16 or FAT32,
+ *  or either.  If the new file system must be FAT32, then query for
+ *  confirmation.  If either filesystem can be used, query for which one.
+ */
+static int
+query_fat_type (PedFileSystem* fs, PedGeometry* new_geom, FatType* out_fat_type)
+{
+	FatSpecific*		fs_info = FAT_SPECIFIC (fs);
+	PedExceptionOption	status;
+	PedSector		fat16_cluster_sectors;
+	PedSector		fat32_cluster_sectors;
+	FatCluster		dummy_cluster_count;
+	PedSector		dummy_fat_sectors;
+	int			fat16_ok;
+	int			fat32_ok;
+
+	fat16_ok = fat_calc_resize_sizes (
+				    new_geom,
+				    fs_info->cluster_sectors,
+				    FAT_TYPE_FAT16,
+				    fs_info->root_dir_sector_count,
+				    fs_info->cluster_sectors,
+				    &fat16_cluster_sectors,
+				    &dummy_cluster_count,
+				    &dummy_fat_sectors);
+
+	fat32_ok = fat_calc_resize_sizes (
+				    new_geom,
+				    fs_info->cluster_sectors,
+				    FAT_TYPE_FAT32,
+				    fs_info->root_dir_sector_count,
+				    fs_info->cluster_sectors,
+				    &fat32_cluster_sectors,
+				    &dummy_cluster_count,
+				    &dummy_fat_sectors);
+
+	if (fat16_ok && fat32_ok) {
+		status = ped_exception_throw (
+			 PED_EXCEPTION_INFORMATION,
+			 PED_EXCEPTION_YES_NO_CANCEL,
+			 _("Would you like to use FAT32?  It is more "
+			 "efficient, but it is not compatible with MS-DOS, "
+			 "early versions of MS-Windows 95 and Windows NT.  If "
+			 "you use these operating systems, then select No.  If "
+			 "you only use Linux, BSD, MS Windows 98 and/or MS "
+			 "Windows 95 B, then you may select Yes."));
+
+		switch (status) {
+		case PED_EXCEPTION_YES:
+			*out_fat_type = FAT_TYPE_FAT32;
+			return 1;
+
+		case PED_EXCEPTION_NO:
+			*out_fat_type = FAT_TYPE_FAT16;
+			return 1;
+
+		case PED_EXCEPTION_CANCEL:
+			return 0;
+		}
+	}
+
+	if (fat16_ok) {
+		*out_fat_type = FAT_TYPE_FAT16;
+		return 1;
+	}
+
+	if (fat32_ok) {
+		*out_fat_type = FAT_TYPE_FAT32;
+		if (fs_info->fat_type == FAT_TYPE_FAT32)
+			return 1;
+
+		status = ped_exception_throw (
+				PED_EXCEPTION_INFORMATION,
+				PED_EXCEPTION_OK_CANCEL,
+				_("The filesystem is going to be too big "
+				"for FAT16, so FAT32 must be used.  This is "
+				"not compatible with MS-DOS, early versions of "
+				"MS-Windows 95 and Windows NT.  If you use "
+				"these operating systems, then select Cancel, "
+				"and create a smaller partition.  If you only "
+				"use Linux, BSD, MS Windows 98 and/or MS "
+				"Windows 95 B, then select OK."));
+
+		return status == PED_EXCEPTION_OK;
+	}
+
+	ped_exception_throw (
+		PED_EXCEPTION_NO_FEATURE,
+		PED_EXCEPTION_CANCEL,
+		_("GNU Parted can not resize this partition to this size.  "
+		"We're working on it!"));
+
+	return 0;
+}
+
 /*  Creates the PedFileSystem struct for the new resized file system, and
     sticks it in a FatOpContext.  At the end of the process, the original
     (ctx->old_fs) is destroyed, and replaced with the new one (ctx->new_fs).
@@ -370,7 +516,7 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	FatSpecific*	new_fs_info;
 	PedFileSystem*	new_fs;
-	int		new_cluster_size;
+	PedSector	new_cluster_sectors;
 	FatCluster	new_cluster_count;
 	PedSector	new_fat_sectors;
 	FatType		new_fat_type;
@@ -386,63 +532,18 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 		root_dir_sector_count = FAT_ROOT_DIR_ENTRY_COUNT
 						* sizeof (FatDirEntry) / 512;
 
-	if (!fat_calc_resize_sizes (new_geom,
-				    fs_info->cluster_sectors,
-				    fs_info->cluster_size,
-				    root_dir_sector_count,
-				    &new_cluster_count,
-				    &new_fat_sectors,
-				    &new_fat_type)) {
-
-		if (!fat_calc_sizes (new_geom, fs_info->cluster_sectors,
-				     FAT_TYPE_FAT16,
-				     root_dir_sector_count,
-				     &new_cluster_size,
-				     &new_cluster_count,
-				     &new_fat_sectors)) {
-			ped_exception_throw (PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_CANCEL,
-				_("Partition too small for a FAT file "
-				  "system."));
-		} else {
-			ped_exception_throw (PED_EXCEPTION_NO_FEATURE,
-				PED_EXCEPTION_CANCEL,
-				_("To resize this filesystem to this "
-				  "size, the cluster size must be changed "
-				  "from %d bytes to %d bytes.  This "
-				  "functionality is not yet implemented.  "
-				  "Note that there are many weird different "
-				  "sizes that can be achieved with the same "
-				  "cluster size - if you keep experimenting "
-				  "with different partition sizes, you might "
-				  "be lucky!!!"),
-				fs_info->cluster_size, new_cluster_size);
-		}
+	if (!query_fat_type (fs, new_geom, &new_fat_type))
 		return 0;
-	}
 
-	if (fs_info->fat_type == FAT_TYPE_FAT16
-			&& new_fat_type == FAT_TYPE_FAT32) {
-		/* FAT32 */
-		if (ped_exception_throw (PED_EXCEPTION_WARNING,
-			PED_EXCEPTION_OK_CANCEL,
-			_("The filesystem is going to be too big for FAT16, "
-			"(with a cluster size of %dk, anyway) so FAT32 will be "
-			"used.  This is not compatible with MS-DOS, early "
-			"versions of MS-Windows 95 and Windows NT.  If you use "
-			"these operating systems, then select cancel, and "
-			"create a smaller partition.  If you only use Linux, "
-			"BSD, MS Windows 98 and/or MS Windows 95 B, then "
-			"select OK."),
-			fs_info->cluster_sectors / 2)
-				== PED_EXCEPTION_CANCEL)
-			return 0;
-	}
+	fat_calc_resize_sizes (new_geom, fs_info->cluster_sectors, new_fat_type,
+		root_dir_sector_count, fs_info->cluster_sectors,
+		&new_cluster_sectors, &new_cluster_count, &new_fat_sectors);
 
-	if (!fat_check_resize_geometry (fs, new_geom, new_cluster_count))
+	if (!fat_check_resize_geometry (fs, new_geom, new_cluster_sectors,
+				        new_cluster_count))
 		goto error;
 
-	new_fs = fat_file_system_alloc (new_geom);
+	new_fs = fat_alloc (new_geom);
 	if (!new_fs)
 		goto error;
 
@@ -459,8 +560,8 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 	new_fs_info->logical_sector_size = fs_info->logical_sector_size;
 	new_fs_info->sector_count = new_geom->length;
 
-	new_fs_info->cluster_size = fs_info->cluster_size;
-	new_fs_info->cluster_sectors = fs_info->cluster_sectors;
+	new_fs_info->cluster_size = new_cluster_sectors * 512;
+	new_fs_info->cluster_sectors = new_cluster_sectors;
 	new_fs_info->cluster_count = new_cluster_count;
 	new_fs_info->dir_entries_per_cluster = fs_info->dir_entries_per_cluster;
 
@@ -486,11 +587,10 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 
 		new_fs_info->cluster_offset
 			= new_fs_info->fat_offset
-			  + 2 * new_fs_info->fat_sectors
-			  - 2 * new_fs_info->cluster_sectors;
+			  + 2 * new_fs_info->fat_sectors;
 	} else {
 		new_fs_info->root_dir_sector_count = root_dir_sector_count;
-		new_fs_info->root_dir_entry_count
+		new_fs_info->root_dir_entry_count 
 			= root_dir_sector_count * 512 / sizeof (FatDirEntry);
 
 		new_fs_info->fat_offset
@@ -501,8 +601,7 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 					       + 2 * new_fs_info->fat_sectors;
 
 		new_fs_info->cluster_offset = new_fs_info->root_dir_offset
-					  + new_fs_info->root_dir_sector_count
-					  - 2 * new_fs_info->cluster_sectors;
+					  + new_fs_info->root_dir_sector_count;
 	}
 	
 	new_fs_info->total_dir_clusters = fs_info->total_dir_clusters;
@@ -514,7 +613,7 @@ create_resize_context (PedFileSystem* fs, PedGeometry* new_geom)
 	if (!fat_op_context_create_initial_fat (context))
 		goto error_free_context;
 
-	if (!fat_file_system_alloc_buffers (new_fs))
+	if (!fat_alloc_buffers (new_fs))
 		goto error_free_fat;
 
 	return context;
@@ -536,6 +635,7 @@ resize_context_assimilate (FatOpContext* ctx)
 {
 	FatSpecific*	old_fs_info = FAT_SPECIFIC (ctx->old_fs);
 
+	fat_free_buffers (ctx->old_fs);
 	fat_table_destroy (old_fs_info->fat);
 	ped_free (old_fs_info);
 	ped_geometry_destroy (ctx->old_fs->geom);
@@ -545,6 +645,24 @@ resize_context_assimilate (FatOpContext* ctx)
 
 	ped_free (ctx->new_fs);
 
+	fat_op_context_destroy (ctx);
+
+	return 1;
+}
+
+static int
+resize_context_abort (FatOpContext* ctx)
+{
+	FatSpecific*	new_fs_info = FAT_SPECIFIC (ctx->new_fs);
+
+	fat_free_buffers (ctx->new_fs);
+	fat_table_destroy (new_fs_info->fat);
+	ped_free (new_fs_info);
+	ped_geometry_destroy (ctx->new_fs->geom);
+	ped_free (ctx->new_fs);
+
+	fat_op_context_destroy (ctx);
+
 	return 1;
 }
 
@@ -558,50 +676,44 @@ fat_resize (PedFileSystem* fs, PedGeometry* geom)
 
 	ctx = create_resize_context (fs, geom);
 	if (!ctx)
-		return 0;
+		goto error;
 	new_fs = ctx->new_fs;
 	new_fs_info = FAT_SPECIFIC (new_fs);
 
 	if (!fat_duplicate_clusters (ctx))
-		return 0;
+		goto error_abort_ctx;
 	if (fs_info->fat_type == FAT_TYPE_FAT16
 			&& new_fs_info->fat_type == FAT_TYPE_FAT32) {
 		if (!alloc_root_dir (ctx))
-			return 0;
+			goto error_abort_ctx;
 	}
 	if (!fat_construct_new_fat (ctx))
-		return 0;
+		goto error_abort_ctx;
 	if (!fat_construct_dir_tree (ctx))
-		return 0;
+		goto error_abort_ctx;
 	if (!fat_table_write_all (new_fs_info->fat, new_fs))
-		return 0;
+		goto error_abort_ctx;
 
 	if (!fat_boot_sector_generate (&new_fs_info->boot_sector, new_fs))
-		return 0;
+		goto error_abort_ctx;
 	if (!fat_boot_sector_write (&new_fs_info->boot_sector, new_fs))
-		return 0;
+		goto error_abort_ctx;
 	if (new_fs_info->fat_type == FAT_TYPE_FAT32) {
 		if (!fat_info_sector_generate (&new_fs_info->info_sector,
 					       new_fs))
-			return 0;
+			goto error_abort_ctx;
 		if (!fat_info_sector_write (&new_fs_info->info_sector, new_fs))
-			return 0;
-		/* I think this is just some boot code...copy it anyway :-) */
-/*
-		if (!ped_geometry_read (fs->geom, fs_info->buffer, 2, 1))
-			return 0;
-		if (!ped_geometry_write (new_fs->geom, fs_info->buffer, 2, 1))
-			return 0;
-*/
+			goto error_abort_ctx;
 	}
 
 	if (!resize_context_assimilate (ctx))
-		return 0;
-
-#ifdef VERBOSE
-	fat_print (fs);
-#endif
+		goto error;
 
 	return 1;
+
+error_abort_ctx:
+	resize_context_abort (ctx);
+error:
+	return 0;
 }
 
