@@ -2,8 +2,8 @@
     libparted - a library for manipulating disk partitions
     Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
 
-	Andrew Clausen			<clausen@alphalink.com.au>
-	Lennert Buytenhek		<buytenh@scotty.dsv.nl>
+	Andrew Clausen			<clausen@gnu.org>
+	Lennert Buytenhek		<buytenh@gnu.org>
 	Matt Wilson, Red Hat Inc.	<msw@redhat.com>
 
     This program is free software; you can redistribute it and/or modify
@@ -21,8 +21,19 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#ifndef LLSEEK_H_INCLUDED
+#define LLSEEK_H_INCLUDED
+
+#include "config.h"
+
+#if SIZEOF_OFF_T < 8 && defined(linux)
+
 #include <sys/types.h>
 
 extern loff_t
 ped_llseek (unsigned int fd, loff_t offset, unsigned int whence);
 
+#endif /*  SIZEOF_OFF_T < 8 && defined(linux) */
+
+#endif /* LLSEEK_H_INCLUDED */
+
