@@ -39,6 +39,30 @@
 #include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
+
+#ifdef HAVE_GETOPT_H
+#include <getopt.h>
+
+static struct option	options[] = {
+	/* name, has-arg, string-return-val, char-return-val */
+	{"help",	0, NULL, 'h'},
+	{"interactive",	0, NULL, 'i'},
+	{"script",	0, NULL, 's'},
+	{"version",	0, NULL, 'v'},
+	{NULL,		0, NULL, 0}
+};
+#endif
+
+static char*	options_help [][2] = {
+	{"help",	N_("displays this help message")},
+	{"interactive",	N_("where necessary, prompts for user intervention")},
+	{"script",	N_("never prompts for user intervention")},
+	{"version",	N_("displays the version")},
+	{NULL,		NULL}
+};
+
+int	opt_script_mode;
 
 static char* minor_msg = N_(
 "MINOR is the partition number used by Linux.  The primary partitions number "
@@ -48,7 +72,7 @@ static char* label_type_msg_start = N_("LABEL-TYPE is one of: ");
 static char* part_type_msg =	N_("PART-TYPE is one of: primary, logical, "
 			           "extended\n");
 static char* fs_type_msg_start = N_("FS-TYPE is one of: ");
-static char* start_end_msg =	N_("START and END are in cylinders\n");
+static char* start_end_msg =	N_("START and END are in megabytes\n");
 static char* device_msg =	N_("DEVICE is usually /dev/hda or /dev/sda\n");
 
 static char* label_type_msg;
@@ -234,7 +258,7 @@ error:
 }
 
 void
-general_help()
+print_commands_help ()
 {
 	int		i;
 
@@ -242,6 +266,18 @@ general_help()
 		command_print_summary (commands [i]);
 }
 
+void
+print_options_help ()
+{
+	int		i;
+
+	for (i=0; options_help [i][0]; i++) {
+		printf ("  -%c, --%-23.23s %s\n",
+			options_help [i][0][0],
+			options_help [i][0],
+			_(options_help [i][1]));
+	}
+}
 
 int
 do_help (PedDevice* dev)
@@ -249,7 +285,7 @@ do_help (PedDevice* dev)
 	if (peek_word ())
 		help_on (get_word ());
 	else
-		general_help();
+		print_commands_help();
 	return 1;
 }
 
@@ -393,13 +429,11 @@ do_mkpart (PedDevice* dev)
 	PedPartition*			part;
 	PedPartitionType		part_type;
 	PedFileSystemType*		fs_type;
-	PedSector			cyl_size;
 	PedSector			start, end;
 
 	disk = ped_disk_open (dev);
 	if (!disk)
 		goto error;
-	cyl_size = dev->heads * dev->sectors;
 
 	if (count_words () < 2) {
 		help_on (_("mkpart"));
@@ -429,21 +463,30 @@ do_mkpart (PedDevice* dev)
 		}
 	}
 
-	if (!is_integer ()) {
+	if (!is_sector ()) {
 		help_on (_("mkpart"));
 		goto error_close_disk;
 	}
-	start = (get_integer() - 1) * cyl_size;
-	if (!is_integer ()) {
+	start = get_sector ();
+	if (!is_sector ()) {
 		help_on (_("mkpart"));
 		goto error_close_disk;
 	}
-	end = (get_integer()) * cyl_size;
+	end = get_sector ();
 	part = ped_partition_new (disk, part_type, fs_type, start, end);
 	if (!part)
 		goto error_close_disk;
 	if (!ped_disk_add_partition (disk, part))
 		goto error_destroy_part;
+
+	/* FIXME:  ped_partition_new() doesn't probe the file system for
+	 * what type of partition should be created.  So we do it here.
+	 * 	Question: SHOULD we probe the file system on
+	 * ped_partition_new()?  It only makes sense if someone accidently
+	 * delete their partition...
+	 */
+       	ped_partition_set_system (part, fs_type);
+
 	ped_disk_write (disk);
 	ped_disk_close (disk);
 	return 1;
@@ -465,13 +508,11 @@ do_mkpartfs (PedDevice* dev)
 	PedPartitionType		part_type;
 	PedFileSystemType*		fs_type;
 	PedFileSystem*			fs;
-	PedSector			cyl_size;
 	PedSector			start, end;
 
 	disk = ped_disk_open (dev);
 	if (!disk)
 		goto error;
-	cyl_size = dev->heads * dev->sectors;
 
 	if (count_words () < 2) {
 		help_on (_("mkpartfs"));
@@ -479,8 +520,8 @@ do_mkpartfs (PedDevice* dev)
 	}
 	switch (get_word() [0]) {
 		case 'p': part_type = PED_PARTITION_PRIMARY; break;
-		case 'e': part_type = PED_PARTITION_EXTENDED; break;
 		case 'l': part_type = PED_PARTITION_LOGICAL; break;
+		case 'e':
 		default: help_on (_("mkpartfs")); goto error_close_disk;
 	}
 
@@ -501,16 +542,16 @@ do_mkpartfs (PedDevice* dev)
 		}
 	}
 
-	if (!is_integer ()) {
+	if (!is_sector ()) {
 		help_on (_("mkpartfs"));
 		goto error_close_disk;
 	}
-	start = (get_integer() - 1) * cyl_size;
-	if (!is_integer ()) {
+	start = get_sector ();
+	if (!is_sector ()) {
 		help_on (_("mkpartfs"));
 		goto error_close_disk;
 	}
-	end = (get_integer()) * cyl_size;
+	end = get_sector ();
 	part = ped_partition_new (disk, part_type, fs_type, start, end);
 	if (!part)
 		goto error_close_disk;
@@ -547,30 +588,27 @@ do_print (PedDevice* dev)
 {
 	PedDisk*	disk;
 	PedPartition*	part;
-	PedSector	cyl_size;
 	int		first_flag;
 
 	disk = ped_disk_open (dev);
 	if (!disk)
 		goto error;
-	cyl_size = dev->heads * dev->sectors;
 
-	printf (_("Disk geometry: 1-%d, %dk cylinders\n"),
-		(int) (disk->dev->length / cyl_size),
-		(int) (cyl_size / 2));
+	printf (_("Disk geometry: 0-%.1f megabytes\n"),
+		disk->dev->length * 1.0 / MEGABYTE_SECTORS);
 
 	printf (
-_("Minor    Start    End      Type            Filesystem     Flags\n"));
+_("Minor   Start     End    Type            Filesystem     Flags\n"));
 
 	for (part = ped_disk_next_partition (disk, NULL); part;
 	     part = ped_disk_next_partition (disk, part)) {
 		if (part->type == PED_PARTITION_PRIMARY
 		    || part->type == PED_PARTITION_LOGICAL
 		    || part->type == PED_PARTITION_EXTENDED) {
-			printf ("%-8d %-8d %-8d %-15s %-15s",
+			printf ("%-5d %8.1f %8.1f  %-15s %-15s",
 				(int) part->num,
-				(int) (1 + part->geom.start / cyl_size),
-				(int) (1 + part->geom.end / cyl_size),
+				(int) part->geom.start * 1.0 / MEGABYTE_SECTORS,
+				(int) part->geom.end * 1.0 / MEGABYTE_SECTORS,
 				ped_partition_type_get_name (part->type),
 				part->fs_type ? part->fs_type->name : "");
 
@@ -612,13 +650,11 @@ do_resize (PedDevice* dev)
 	PedPartition*		part;
 	PedFileSystem*		fs;
 	PedFileSystemType*	fs_type;
-	PedSector		cyl_size;
 	PedSector		start, end;
 
 	disk = ped_disk_open (dev);
 	if (!disk)
 		goto error;
-	cyl_size = dev->heads * dev->sectors;
 
 	if (!is_integer ()) {
 		help_on (_("resize"));
@@ -639,16 +675,18 @@ do_resize (PedDevice* dev)
 			goto error_close_disk;
 	}
 
-	if (!is_integer ()) {
+	if (!is_sector ()) {
 		help_on (_("resize"));
 		goto error_close_disk;
 	}
-	start = (get_integer() - 1) * cyl_size;
-	if (!is_integer ()) {
+	start = get_sector ();
+	if (!is_sector ()) {
 		help_on (_("resize"));
 		goto error_close_disk;
 	}
-	end = (get_integer()) * cyl_size;
+
+	/* NOTE: this gets aligned by ped_disk_set_partition_geom() */
+	end = get_sector ();
 
 	if (part->type == PED_PARTITION_EXTENDED) {
 		if (!ped_disk_set_partition_geom (disk, part, start, end))
@@ -754,8 +792,8 @@ error:
 	return 0;
 }
 
-void
-init_messages ()
+static void
+_init_messages ()
 {
 	StrList*		list;
 	int			first;
@@ -797,8 +835,8 @@ init_messages ()
 	str_list_destroy (list);
 }
 
-void
-init_commands ()
+static void
+_init_commands ()
 {
 	command_register (commands, command_create (
 		str_list_create_unique ("boot", _("boot"), NULL),
@@ -846,7 +884,7 @@ NULL),
 		str_list_create (
 _("mklabel LABEL-TYPE            create a new disklabel (partition table)"),
 NULL),
-		str_list_create (_(label_type_msg), NULL)));
+		str_list_create (label_type_msg, NULL)));
 
 	command_register (commands, command_create (
 		str_list_create_unique ("mkfs", _("mkfs"), NULL),
@@ -925,8 +963,8 @@ NULL),
 
 }
 
-int
-init ()
+static void
+_init_i18n ()
 {
 /* intialize i18n */
 #ifdef ENABLE_NLS
@@ -935,60 +973,131 @@ init ()
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 #endif /* ENABLE_NLS */
+}
 
-	if (!init_ui ())
-		return 0;
+void
+_version ()
+{
+	printf (prog_name);
+	exit (0);
+}
 
-	if (!ped_init ())
-		return 0;
+static int
+_parse_options (int* argc_ptr, char*** argv_ptr)
+{
+	int	opt;
 
-	init_messages ();
-	init_commands ();
+	while (1)
+	{
+#ifdef HAVE_GETOPT_H
+		opt = getopt_long (*argc_ptr, *argv_ptr, "hisv",
+				   options, NULL);
+#else
+		opt = getopt (*argc_ptr, *argv_ptr, "hisv");
+#endif
+		if (opt == -1)
+			break;
 
+		switch (opt) {
+			case 'h': help_msg (); break;
+			case 'i': opt_script_mode = 0; break;
+			case 's': opt_script_mode = 1; break;
+			case 'v': _version (); break;
+		}
+	}
+
+	*argc_ptr -= optind;
+	*argv_ptr += optind;
 	return 1;
+
+error:
+	return 0;
+}
+
+static PedDevice*
+_choose_device (int* argc_ptr, char*** argv_ptr)
+{
+	PedDevice*	dev;
+
+	/* specified on comand line? */
+	if (*argc_ptr) {
+		dev = ped_device_get ((*argv_ptr) [0]);
+		if (!dev)
+			return NULL;
+		(*argc_ptr)--;
+		(*argv_ptr)++;
+	} else {
+	retry:
+		ped_device_probe_all ();
+		dev = ped_device_get_next (NULL);
+		if (!dev) {
+			if (ped_exception_throw (PED_EXCEPTION_ERROR,
+				PED_EXCEPTION_RETRY_CANCEL,
+				_("No device found"))
+					== PED_EXCEPTION_RETRY)
+				goto retry;
+			else
+				return NULL;
+		}
+	}
+
+	if (!ped_device_open (dev))
+		return NULL;
+	return dev;	
+}
+
+static PedDevice*
+_init (int* argc_ptr, char*** argv_ptr)
+{
+	PedDevice*	dev;
+
+	_init_i18n ();
+	if (!init_ui ())
+		goto error;
+	if (!ped_init ())
+		goto error;
+	_init_messages ();
+	_init_commands ();
+
+	if (!_parse_options (argc_ptr, argv_ptr))
+		goto error;
+	dev = _choose_device (argc_ptr, argv_ptr);
+	if (!dev)
+		goto error;
+
+	return dev;
+
+error_done_ui:
+	done_ui ();
+error:
+	return NULL;
+}
+
+static void
+_done (PedDevice *dev)
+{
+	ped_device_close (dev);
+	ped_done ();
+	done_ui ();
 }
 
 int
 main (int argc, char** argv)
 {
-	PedDevice*	dev = NULL;
-	Command*	cmd;
+	PedDevice*	dev;
 	int		status;
 
-	if (!init ())
-		goto error;
+	dev = _init (&argc, &argv);
+	if (!dev)
+		return 0;
 
-	while (1) {
-		if (argc > 1) {
-			dev = ped_device_get (argv [1]);
-			if (!dev)
-				goto error;
-		} else {
-			ped_device_probe_all ();
-			dev = ped_device_get_next (NULL);
-		}
-		if (dev)
-			break;
-		if (ped_exception_throw (PED_EXCEPTION_ERROR,
-			PED_EXCEPTION_RETRY_CANCEL,
-			_("No device found"))
-				== PED_EXCEPTION_CANCEL)
-			goto error;
-	}
-	if (!ped_device_open (dev))
-		goto error;
-
-	if (argc > 2)
+	if (argc)
 		status = non_interactive_mode (dev, commands, argc, argv);
 	else
 		status = interactive_mode (dev, commands);
 
-	ped_device_close (dev);
-	ped_done ();
-	done_ui ();
+	_done (dev);
 
 	return !status;
-
-error:
-	return 1;
 }
+
