@@ -23,6 +23,8 @@
 
 #include "config.h"
 
+#include <parted/debug.h>
+
 #include <ctype.h>
 #include <errno.h>
 #include <stdarg.h>
@@ -453,11 +455,11 @@ str_search (const wchar_t* str, int n, wchar_t c)
 
 #ifdef ENABLE_NLS
 	for (i=0; i<n; i++)
-		if (str [i] == c)
+		if (str [i] == c << (3 * 8))
 			return i;
 #else
 	for (i=0; i<n; i++)
-		if (str [i] == c << (3 * 8))
+		if (str [i] == c)
 			return i;
 #endif
 	return -1;
@@ -470,6 +472,7 @@ str_search (const wchar_t* str, int n, wchar_t c)
  * need to include them.  Best not to avoid side effects, like 3.
 14159 :-)
  * FIXME: how do we exclude "." and "(" ?
+ * FIXME: glibc doesn't like umlaute.  i.e. \"o (TeX equivalent)
  */
 
 static int
@@ -502,13 +505,15 @@ str_list_print_wrap (const StrList* list, int line_length, int offset,
 	const StrList*	walk;
 	const wchar_t*	str;
 	int		str_len;
-	unsigned int	cut_right;
-	unsigned int	cut_left;
-	unsigned int	line_left;
+	int		cut_right;
+	int		cut_left;
+	int		line_left;
 	char*		spaces;
 	int		search_result;
 	int		line_break;
 
+	PED_ASSERT (line_length - indent > 10, return);
+
 	spaces = get_spaces (indent);
 	line_left = line_length - offset;
 
@@ -518,8 +523,16 @@ str_list_print_wrap (const StrList* list, int line_length, int offset,
 		str = walk->str;
 		str_len = wchar_strlen (str);
 
-		while (str_len > line_left || wchar_strchr (str, '\n')) {
+		while (line_left < str_len || wchar_strchr (str, '\n')) {
+			line_break = 0;
+
 			cut_left = MIN (line_left - 1, str_len - 1);
+
+			/* we can have a space "over", but not a comma */
+			if (cut_left < str_len
+					&& is_space (str [cut_left + 1]))
+				cut_left++;
+
 			while (cut_left && !is_break_point (str [cut_left]))
 				cut_left--;
 			while (cut_left && is_space (str [cut_left]))
@@ -529,15 +542,18 @@ str_list_print_wrap (const StrList* list, int line_length, int offset,
 		 * Japanese character, or the start of a blank line.
 		 */
 
-			search_result = str_search (str, cut_left, '\n');
-			if (search_result != -1)
+			search_result = str_search (str, cut_left + 1, '\n');
+			if (search_result != -1) {
 				cut_left = search_result - 1;
+				line_break = 1;
+			}
 
-			for (cut_right = cut_left + 1;
+			for (cut_right = cut_left + 2;
 			     cut_right < str_len && is_space (str [cut_right]);
 			     cut_right++);
 
-			print_wchar (str, cut_left + 1);
+			if (cut_left > 0)
+				print_wchar (str, cut_left + 1);
 
 			str += cut_right;
 			str_len -= cut_right;
@@ -545,6 +561,8 @@ str_list_print_wrap (const StrList* list, int line_length, int offset,
 
 			if (walk->next || *str)
 				printf ("\n%s", spaces);
+			else if (line_break)
+				printf ("\n");
 		}
 
 		print_wchar (str, 0);
