@@ -24,6 +24,11 @@
 #ifndef UI_H_INCLUDED
 #define UI_H_INCLUDED
 
+#define MEGABYTE (1024 * 1024)
+#define MEGABYTE_SECTORS (MEGABYTE / 512)
+
+extern char*	prog_name;
+
 extern int init_ui ();
 extern int non_interactive_mode (PedDevice* dev, Command* cmd_list[],
 				 int argc, char* argv[]);
@@ -34,13 +39,20 @@ extern int screen_width ();
 
 extern char* get_word ();
 extern int get_integer ();
+extern PedSector get_sector ();
 extern char* peek_word ();
 extern int is_integer ();
+extern int is_sector ();
 extern int count_words ();
 
+extern void help_msg ();
+
 /* in parted.c */
+extern int	opt_script_mode;
+
 extern int do_help (PedDevice* dev);
-extern void general_help ();
+extern void print_options_help ();
+extern void print_commands_help ();
 
 #endif /* UI_H_INCLUDED */
 
